# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack.stacks import _nova


CONF = config.CONF

UBUNTU_IMAGE_VERSION = 'focal'

UBUNTU_IMAGE_VERSION_NUMBER = '20.04'

UBUNTU_MINIMAL_IMAGE_URL = (
    'https://cloud-images.ubuntu.com/minimal/releases/'
    f'{UBUNTU_IMAGE_VERSION}/release/'
    f'ubuntu-{UBUNTU_IMAGE_VERSION_NUMBER}-minimal-cloudimg-amd64.img')


class UbuntuMinimalImageFixture(glance.URLGlanceImageFixture):
    image_url = CONF.tobiko.ubuntu.image_url or UBUNTU_MINIMAL_IMAGE_URL
    image_name = CONF.tobiko.ubuntu.image_name
    image_file = CONF.tobiko.ubuntu.image_file
    disk_format = CONF.tobiko.ubuntu.disk_format or "qcow2"
    container_format = CONF.tobiko.ubuntu.container_format or "bare"
    username = CONF.tobiko.ubuntu.username or 'ubuntu'
    password = CONF.tobiko.ubuntu.password
    connection_timeout = CONF.tobiko.ubuntu.connection_timeout or 600.


IPERF3_SERVICE_FILE = """
[Unit]
Description=iperf3 server on port %i
After=syslog.target network.target

[Service]
ExecStart=/usr/bin/iperf3 -s -p %i
Restart=always
RuntimeMaxSec=3600
User=root

[Install]
WantedBy=multi-user.target
DefaultInstance=5201
"""


class UbuntuImageFixture(UbuntuMinimalImageFixture,
                         glance.CustomizedGlanceImageFixture):
    """Ubuntu server image running an HTTP server

    The server has additional installed packages compared to
    the minimal one:
      - iperf3
      - ping
      - ncat
      - nginx

    The image will also have below running services:
      - nginx HTTP server listening on TCP port 80
      - iperf3 server listening on TCP port 5201
    """

    @property
    def firstboot_commands(self) -> typing.List[str]:
        return super().firstboot_commands + [
            'sh -c "hostname > /var/www/html/id"']

    @property
    def install_packages(self) -> typing.List[str]:
        return super().install_packages + ['iperf3',
                                           'iputils-ping',
                                           'ncat',
                                           'nginx']

    # port of running HTTP server
    http_port = 80

    # port of running Iperf3 server
    iperf3_port = 5201

    @property
    def run_commands(self) -> typing.List[str]:
        run_commands = super().run_commands
        run_commands.append(
            f'echo "{IPERF3_SERVICE_FILE}" '
            '> /etc/systemd/system/iperf3-server@.service')
        run_commands.append(
            f"systemctl enable iperf3-server@{self.iperf3_port}")
        return run_commands


class UbuntuFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 128
    swap = 512


class UbuntuMinimalServerStackFixture(_nova.CloudInitServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_setup_fixture(UbuntuMinimalImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_setup_fixture(UbuntuFlavorStackFixture)


class UbuntuServerStackFixture(UbuntuMinimalServerStackFixture):
    """Ubuntu server running an HTTP server

    The server has additional commands compared to the minimal one:
      iperf3
      ping
    """

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_setup_fixture(UbuntuImageFixture)

    # port of running HTTP server
    @property
    def http_port(self) -> int:
        return self.image_fixture.http_port

    @property
    def iperf3_port(self) -> int:
        return self.image_fixture.iperf3_port


class UbuntuExternalServerStackFixture(UbuntuServerStackFixture,
                                       _nova.ExternalServerStackFixture):
    """Ubuntu server with port on special external network
    """



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
import testtools

import tobiko
from tobiko.shell import ifconfig
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import keystone
from tobiko.openstack import stacks


class IfconfigTest(testtools.TestCase):

    cirros_stack = tobiko.required_setup_fixture(
        stacks.CirrosServerStackFixture)

    ubuntu_stack = tobiko.required_setup_fixture(
        stacks.UbuntuServerStackFixture)

    def test_list_ip_addresses(self, ip_version=None, **execute_params):
        result = sh.execute(command='ls /sbin/ifconfig',
                            expect_exit_status=None, **execute_params)
        if result.exit_status != 0:
            self.skip(f"{result.stderr}")

        ips = ifconfig.list_ip_addresses(ip_version=ip_version,
                                         **execute_params)
        self.assertIsInstance(ips, tobiko.Selection)
        for ip in ips:
            self.assertIsInstance(ip, netaddr.IPAddress)
        if ip_version:
            self.assertEqual(ips.with_attributes(version=ip_version), ips)

    def test_list_ip_addresses_with_ipv4(self):
        self.test_list_ip_addresses(ip_version=4)

    def test_list_ip_addresses_with_ipv6(self):
        self.test_list_ip_addresses(ip_version=6)

    @keystone.skip_unless_has_keystone_credentials()
    def test_list_ip_addresses_with_cirros_server(self):
        self.test_list_ip_addresses(ssh_client=self.cirros_stack.ssh_client)

    @keystone.skip_unless_has_keystone_credentials()
    def test_list_ip_addresses_with_ubuntu_server(self):
        self.test_list_ip_addresses(ssh_client=self.ubuntu_stack.ssh_client)

    def test_list_ip_addresses_with_proxy_ssh_client(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is None:
            self.skip('SSH proxy server not configured')
        self.test_list_ip_addresses(ssh_client=ssh_client)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack.stacks import _centos
from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _hot
from tobiko.openstack.stacks import _neutron


CONF = config.CONF
LOG = log.getLogger(__name__)


class OctaviaVipNetworkStackFixture(_neutron.NetworkStackFixture):
    # Load Balancer VIP network must use port security (required by neutron to
    # support allowed address pairs on ports)
    port_security_enabled = True


class OctaviaMemberNetworkStackFixture(_neutron.NetworkStackFixture):
    pass


class OctaviaCentosServerStackFixture(_centos.CentosServerStackFixture):
    network_stack = tobiko.required_setup_fixture(
        OctaviaMemberNetworkStackFixture)

    @property
    def user_data(self):
        # Launch a webserver on port 80 that replies the server name to the
        # client
        return ("#cloud-config\n"
                "packages:\n"
                "- httpd\n"
                "runcmd:\n"
                "- [ sh, -c, \"hostname > /var/www/html/id\" ]\n"
                "- [ systemctl, enable, --now, httpd ]\n")


class OctaviaCirrosServerStackFixture(_cirros.CirrosServerStackFixture):
    network_stack = tobiko.required_setup_fixture(
        OctaviaMemberNetworkStackFixture)

    @property
    def user_data(self):
        # Launch a webserver on port 80 that replies the server name to the
        # client
        # This webserver relies on the nc command which may fail if multiple
        # clients connect at the same time. For concurrency testing,
        # OctaviaCentosServerStackFixture is more suited to handle multiple
        # requests.

        return (
            "#!/bin/sh\n"
            "sudo nc -k -p 80 -e echo -e \"HTTP/1.1 200 OK\r\n"
            "Content-Length: $(hostname | head -c-1 | wc -c )\r\n"
            "Server: $(hostname)\r\n"
            "Content-type: text/html; charset=utf-8\r\n"
            "Connection: close\r\n\r\n"
            "$(hostname)\"\n")


class OctaviaServerStackFixture(OctaviaCirrosServerStackFixture):
    pass


class OctaviaLoadbalancerStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/load_balancer.yaml')

    vip_network = tobiko.required_setup_fixture(OctaviaVipNetworkStackFixture)

    ip_version = 4

    @property
    def vip_subnet_id(self):
        if self.ip_version == 4:
            return self.vip_network.ipv4_subnet_id
        else:
            return self.vip_network.ipv6_subnet_id


class OctaviaListenerStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/listener.yaml')

    loadbalancer = tobiko.required_setup_fixture(
        OctaviaLoadbalancerStackFixture)

    lb_port = 80

    lb_protocol = 'HTTP'

    @property
    def loadbalancer_id(self):
        return self.loadbalancer.loadbalancer_id


class OctaviaPoolStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/pool.yaml')

    listener = tobiko.required_setup_fixture(
        OctaviaListenerStackFixture)

    pool_protocol = 'HTTP'

    lb_algorithm = 'ROUND_ROBIN'

    hm_type = 'HTTP'

    # healthmonitor attributes
    hm_delay = 3

    hm_max_retries = 4

    hm_timeout = 3

    hm_type = 'HTTP'

    @property
    def listener_id(self):
        return self.listener.listener_id


class OctaviaMemberServerStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('octavia/member.yaml')

    pool = tobiko.required_setup_fixture(OctaviaPoolStackFixture)

    server_stack = tobiko.required_setup_fixture(OctaviaServerStackFixture)

    application_port = 80

    ip_version = 4

    @property
    def pool_id(self):
        return self.pool.pool_id

    @property
    def subnet_id(self):
        if self.ip_version == 4:
            return self.server_stack.network_stack.ipv4_subnet_id
        else:
            return self.server_stack.network_stack.ipv6_subnet_id

    @property
    def member_address(self):
        return [
            fixed_ip['ip_address']
            for fixed_ip in self.server_stack.fixed_ips
            if ((self.ip_version == 4 and
                 ':' not in fixed_ip['ip_address']) or
                (self.ip_version == 6 and
                 ':' in fixed_ip['ip_address']))
        ][0]


class OctaviaClientServerStackFixture(_cirros.CirrosServerStackFixture):
    network_stack = tobiko.required_setup_fixture(
        OctaviaVipNetworkStackFixture)


class OctaviaOtherServerStackFixture(
        OctaviaServerStackFixture):
    pass


class OctaviaOtherMemberServerStackFixture(
        OctaviaMemberServerStackFixture):
    server_stack = tobiko.required_setup_fixture(
        OctaviaOtherServerStackFixture)



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class LocalPsTest(testtools.TestCase):

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return False

    def test_list_processes(self):
        processes = sh.list_processes(ssh_client=self.ssh_client)
        self._check_processes(processes,
                              is_kernel=False)

    def test_list_kernel_processes(self):
        processes = sh.list_kernel_processes(ssh_client=self.ssh_client)
        self._check_processes(processes=processes, is_kernel=True)

    def test_list_all_processes(self):
        processes = sh.list_all_processes(ssh_client=self.ssh_client)
        self._check_processes(processes=processes, is_kernel=None)

    def test_list_processes_with_pid(self):
        processes = sh.list_processes(ssh_client=self.ssh_client)
        processes_with_pid = sh.list_processes(pid=processes[0].pid,
                                               ssh_client=self.ssh_client)
        self.assertEqual(processes[:1], processes_with_pid)

    def test_list_processes_with_command(self):
        processes = sh.list_processes(command='systemd',
                                      ssh_client=self.ssh_client)
        for process in processes:
            self.assertTrue(process.command.startswith('systemd'), process)

    def test_list_processes_with_command_line(self):
        cat_process = sh.process('cat -',
                                 ssh_client=self.ssh_client).execute()
        self.addCleanup(cat_process.kill)
        processes = sh.list_processes(command_line='cat -',
                                      ssh_client=self.ssh_client)
        for process in processes:
            self.assertEqual('cat', process.command)
            self.assertEqual(('cat', '-'), process.command_line)
        cat_process.kill()
        sh.wait_for_processes(command_line='cat -',
                              timeout=30.,
                              ssh_client=self.ssh_client)

    def test_list_processes_with_exact_command(self):
        processes = sh.list_processes(command='^systemd$',
                                      ssh_client=self.ssh_client)
        self.assertEqual(processes.with_attributes(command='systemd'),
                         processes)

    def _check_processes(self, processes, is_kernel):
        self.assertIsInstance(processes, tobiko.Selection)
        for process in processes:
            self.assertGreater(process.pid, 0)
            self.assertIs(
                (process.command.startswith('[') and
                 process.command.endswith(']')),
                process.is_kernel)
            if is_kernel is not None:
                self.assertIs(bool(is_kernel), process.is_kernel)

    def test_wait_for_processes(self):
        # assume the PID of the first execution of PS process is not more there
        # at the second execution
        process = sh.list_processes(command='ps',
                                    ssh_client=self.ssh_client)[-1]
        sh.wait_for_processes(pid=process.pid,
                              command='ps',
                              timeout=30.,
                              ssh_client=self.ssh_client)

    def test_wait_for_processes_timeout(self):
        # assume there are always to be running processes on host
        ex = self.assertRaises(sh.PsWaitTimeout, sh.wait_for_processes,
                               timeout=3.,
                               ssh_client=self.ssh_client)
        self.assertEqual(3., ex.timeout)
        self.assertEqual(sh.get_hostname(ssh_client=self.ssh_client),
                         ex.hostname)


class CirrosPsTest(LocalPsTest):

    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return self.stack.ssh_client


class SSHPsTest(LocalPsTest):

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        ssh_client = ssh.ssh_proxy_client()
        if isinstance(ssh_client, ssh.SSHClientFixture):
            return ssh_client

        nodes = topology.list_openstack_nodes()
        for node in nodes:
            if isinstance(node.ssh_client, ssh.SSHClientFixture):
                return ssh_client
        tobiko.skip_test('No such SSH server host to connect to')



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.tests import unit


LOG = log.getLogger(__name__)


class DefaultKeystoneCredentialsPatch(unit.PatchFixture):

    credentials = keystone.keystone_credentials(
        auth_url='http://127.0.0.1:5000/v3',
        username='default',
        project_name='default',
        password='this is a secret')

    def __init__(self, credentials=None):
        if credentials:
            self.credentials = credentials
        tobiko.check_valid_type(self.credentials,
                                keystone.KeystoneCredentials)

    def setup_fixture(self):
        CONF = config.CONF
        keystone_conf = CONF.tobiko.keystone
        for key, value in self.credentials.to_dict().items():
            self.patch(keystone_conf, key, value)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import ironicclient.v1.client
from oslo_log import log

import tobiko
from tobiko.openstack.ironic import _client


LOG = log.getLogger(__name__)

IronicNode = typing.Union[ironicclient.v1.node.Node]
IronicNodeType = typing.Union[str, IronicNode]


def get_node_id(node: typing.Optional[IronicNodeType] = None,
                node_id: typing.Optional[str] = None) -> str:
    if node_id is None:
        if isinstance(node, str):
            node_id = node
        else:
            assert node is not None
            node_id = node.uuid
    return node_id


def get_node(node: typing.Optional[IronicNodeType] = None,
             node_id: typing.Optional[str] = None,
             client: _client.IronicClientType = None,
             **params) -> IronicNode:
    node_id = get_node_id(node=node, node_id=node_id)
    return _client.ironic_client(client).node.get(node_id, **params)


class WaitForNodePowerStateError(tobiko.TobikoException):
    message = ("Node {node_id} not changing power state from "
               "{node_power_state} to {power_state}")


class WaitForNodePowerStateTimeout(WaitForNodePowerStateError):
    message = ("Node {node_id} didn't change its state from "
               "{node_power_state} to {power_state} state after "
               "{timeout} seconds")


IRONIC_NODE_TRANSIENT_POWER_STATES: typing.Dict[str, typing.List[str]] = {
    'power on': ['power off'],
    'power off': ['power on'],
}


def wait_for_node_power_state(
        node: IronicNodeType,
        power_state: str,
        client: _client.IronicClientType = None,
        timeout: tobiko.Seconds = None,
        sleep_time: tobiko.Seconds = None,
        transient_status: typing.Optional[typing.List[str]] = None) -> \
            IronicNode:
    if transient_status is None:
        transient_status = IRONIC_NODE_TRANSIENT_POWER_STATES.get(
            power_state) or []
    node_id = get_node_id(node)
    for attempt in tobiko.retry(timeout=timeout,
                                interval=sleep_time,
                                default_timeout=300.,
                                default_interval=5.):
        _node = get_node(node_id=node_id, client=client)
        if _node.power_state == power_state:
            break
        if _node.power_state not in transient_status:
            raise WaitForNodePowerStateError(
                node_id=node_id,
                node_power_state=_node.power_state,
                power_state=power_state)
        if attempt.is_last:
            raise WaitForNodePowerStateTimeout(
                node_id=node_id,
                node_power_state=_node.power_state,
                power_state=power_state,
                timeout=timeout)

        LOG.debug(f"Waiting for Ironic node '{node_id}' power state to get "
                  f"from {_node.power_state} to {power_state}...")
    else:
        raise RuntimeError("Retry look break before timing out")

    return _node


def power_off_node(node: IronicNodeType,
                   soft=False,
                   client: _client.IronicClientType = None,
                   timeout: tobiko.Seconds = None,
                   sleep_time: tobiko.Seconds = None) \
        -> IronicNode:
    client = _client.ironic_client(client)
    node = get_node(node=node, client=client)
    if node.power_state == 'power off':
        return node

    LOG.info(f"Power off baremetal node '{node.uuid}' "
             f"(power state = '{node.power_state}').")
    client.node.set_power_state(node.uuid,
                                state='off',
                                soft=soft,
                                timeout=timeout)
    return wait_for_node_power_state(node=node.uuid,
                                     power_state='power off',
                                     client=client,
                                     timeout=timeout,
                                     sleep_time=sleep_time)


def power_on_node(node: IronicNodeType,
                  client: _client.IronicClientType = None,
                  timeout: tobiko.Seconds = None,
                  sleep_time: tobiko.Seconds = None) -> \
        IronicNode:
    client = _client.ironic_client(client)
    node = get_node(node=node, client=client)
    if node.power_state == 'power on':
        return node

    LOG.info(f"Power on baremetal node '{node.uuid}' "
             f"(power_state='{node.power_state}').")
    client.node.set_power_state(node_id=node.uuid, state='on')

    return wait_for_node_power_state(node=node.uuid,
                                     power_state='power on',
                                     client=client,
                                     timeout=timeout,
                                     sleep_time=sleep_time)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


from oslo_log import log
from neutron_lib import constants

import tobiko
from tobiko.shell.ping import _exception
from tobiko.shell.ping import _parameters
from tobiko.shell import sh


LOG = log.getLogger(__name__)


def get_ping_command(parameters, ssh_client):
    interface = get_ping_interface(ssh_client=ssh_client)
    return interface.get_ping_command(parameters)


def get_ping_interface(ssh_client):
    manager = tobiko.setup_fixture(PingInterfaceManager)
    interface = manager.get_ping_interface(ssh_client=ssh_client)
    tobiko.check_valid_type(interface, PingInterface)
    return interface


def has_fragment_ping_option(ssh_client=None):
    interface = get_ping_interface(ssh_client=ssh_client)
    return interface.has_fragment_option


skip_if_missing_fragment_ping_option = tobiko.skip_unless(
    "requires (don't) fragment Ping option",
    has_fragment_ping_option)


class PingInterfaceManager(tobiko.SharedFixture):

    def __init__(self):
        super(PingInterfaceManager, self).__init__()
        self.client_interfaces = {}
        self.interfaces = []
        self.default_interface = PingInterface()

    def add_ping_interface(self, interface):
        LOG.debug('Register ping interface %r', interface)
        self.interfaces.append(interface)

    def get_ping_interface(self, ssh_client):
        try:
            return self.client_interfaces[ssh_client]
        except KeyError:
            pass

        LOG.debug('Looking for ping interface for SSH client %s', ssh_client)
        usage = get_ping_usage(ssh_client)
        interface = find_ping_interface(usage=usage,
                                        interfaces=self.interfaces)
        if not interface:
            LOG.error('Ping interface not found: using the default one')
            interface = self.default_interface
        LOG.debug('Assign Ping interface %r to SSH client %r',
                  interface, ssh_client)
        self.client_interfaces[ssh_client] = interface
        return interface


def get_ping_usage(ssh_client):
    result = sh.execute('ping --help', expect_exit_status=None,
                        ssh_client=ssh_client)
    usage = ((result.stdout and str(result.stdout)) or
             (result.stderr and str(result.stderr)) or "").strip()
    if result.exit_status != 0 and 'command not found' in usage.lower():
        raise tobiko.SkipException(
            'ping command not installed on this instance')
    if usage:
        LOG.debug('Got ping usage text')
    else:
        raise ValueError("Unable to get usage message from ping command:\n"
                         "%s" % result.details)
    return usage


def find_ping_interface(usage, interfaces):
    if usage:
        for interface in interfaces:
            if interface.match_ping_usage(usage):
                return interface

    LOG.warning("No such ping interface class from usage message:\n"
                "%r", usage)
    return None


def ping_interface(interface_class):
    assert issubclass(interface_class, PingInterface)
    manager = tobiko.setup_fixture(PingInterfaceManager)
    manager.add_ping_interface(interface=interface_class())
    return interface_class


class PingInterface(object):

    ping_usage = None

    def match_ping_usage(self, usage):
        # pylint: disable=unused-argument
        return False

    def get_ping_command(self, parameters):
        host = parameters.host
        if not host:
            raise ValueError(f"Invalid destination host: '{host}'")

        command = sh.shell_command([self.get_ping_executable(parameters)] +
                                   self.get_ping_options(parameters) +
                                   [host])
        LOG.debug('Got ping command from interface %r for host %r: %r',
                  self, host, command)
        return command

    def get_ping_executable(self, parameters):
        ip_version = _parameters.get_ping_ip_version(parameters)
        if ip_version == constants.IP_VERSION_6:
            return 'ping6'
        else:
            return 'ping'

    def get_ping_options(self, parameters):
        options = []

        ip_version = _parameters.get_ping_ip_version(parameters)
        if ip_version == constants.IP_VERSION_4:
            options += self.get_ipv4_option()
        elif ip_version == constants.IP_VERSION_6:
            options += self.get_ipv6_option()
        elif ip_version is not None:
            message = 'Invalid IP version: {!r}'.format(ip_version)
            raise ValueError(message)

        interface = parameters.source
        if interface:
            options += self.get_interface_option(interface)

        deadline = parameters.deadline
        if deadline > 0:
            options += self.get_deadline_option(parameters)

        count = parameters.count
        if count > 0:
            options += self.get_count_option(count)

        size = _parameters.get_ping_payload_size(parameters)
        if size:
            options += self.get_size_option(size)

        interval = parameters.interval
        if interval > 1:
            options += self.get_interval_option(interval)

        fragment = parameters.fragmentation
        if fragment is not None:
            options += self.get_fragment_option(fragment=fragment)

        return options

    def get_ipv4_option(self):
        return []

    def get_ipv6_option(self):
        return []

    def get_interface_option(self, interface):
        return ['-I', interface]

    def get_deadline_option(self, parameters):
        return ['-w', parameters.deadline,
                '-W', parameters.deadline]

    def get_count_option(self, count):
        return ['-c', int(count)]

    def get_size_option(self, size):
        return ['-s', int(size)]

    def get_interval_option(self, interval):
        return ['i', int(interval)]

    has_fragment_option = False

    def get_fragment_option(self, fragment):
        details = ("{!r} ping implementation doesn't support "
                   "'fragment={!r}' option").format(self, fragment)
        raise _exception.UnsupportedPingOption(details=details)


class IpVersionPingInterface(PingInterface):

    def get_ping_executable(self, parameters):
        # pylint: disable=unused-argument
        return 'ping'

    def get_ipv4_option(self):
        return ['-4']

    def get_ipv6_option(self):
        return ['-6']


IPUTILS_PING_USAGE = """
ping: invalid option -- '-'
Usage: ping [-aAbBdDfhLnOqrRUvV] [-c count] [-i interval] [-I interface]
            [-m mark] [-M pmtudisc_option] [-l preload] [-p pattern] [-Q tos]
            [-s packetsize] [-S sndbuf] [-t ttl] [-T timestamp_option]
            [-w deadline] [-W timeout] [hop1 ...] destination
""".strip()


@ping_interface
class IpUtilsPingInterface(PingInterface):

    def match_ping_usage(self, usage):
        return usage.startswith(IPUTILS_PING_USAGE)

    has_fragment_option = True

    def get_fragment_option(self, fragment):
        if fragment:
            return ['-M', 'dont']
        else:
            return ['-M', 'do']


IP_VERSION_IPUTILS_PING_USAGE = """
ping: invalid option -- '-'
Usage: ping [-aAbBdDfhLnOqrRUvV64] [-c count] [-i interval] [-I interface]
            [-m mark] [-M pmtudisc_option] [-l preload] [-p pattern] [-Q tos]
            [-s packetsize] [-S sndbuf] [-t ttl] [-T timestamp_option]
            [-w deadline] [-W timeout] [hop1 ...] destination
Usage: ping -6 [-aAbBdDfhLnOqrRUvV] [-c count] [-i interval] [-I interface]
             [-l preload] [-m mark] [-M pmtudisc_option]
             [-N nodeinfo_option] [-p pattern] [-Q tclass] [-s packetsize]
             [-S sndbuf] [-t ttl] [-T timestamp_option] [-w deadline]
             [-W timeout] destination
""".strip()


@ping_interface
class IpUtilsIpVersionPingInterface(IpUtilsPingInterface,
                                    IpVersionPingInterface):

    def match_ping_usage(self, usage):
        return usage.startswith(IP_VERSION_IPUTILS_PING_USAGE)


BUSYBOX_PING_USAGE = """
ping: unrecognized option `--usage'
BusyBox v1.23.2 (2017-11-20 02:37:12 UTC) multi-call binary.

Usage: ping [OPTIONS] HOST

Send ICMP ECHO_REQUEST packets to network hosts

        -4,-6          Force IP or IPv6 name resolution
        -c CNT         Send only CNT pings
        -s SIZE        Send SIZE data bytes in packets (default:56)
        -t TTL         Set TTL
        -I IFACE/IP    Use interface or IP address as source
        -W SEC         Seconds to wait for the first response (default:10)
                       (after all -c CNT packets are sent)
        -w SEC         Seconds until ping exits (default:infinite)
                       (can exit earlier with -c CNT)
        -q             Quiet, only display output at start
                       and when finished
        -p             Pattern to use for payload
""".strip()


@ping_interface
class BusyBoxPingInterface(IpVersionPingInterface):

    def match_ping_usage(self, usage):
        return usage.startswith("BusyBox")


INET_TOOLS_PING_USAGE = """
Usage: ping [OPTION...] HOST ...
Send ICMP ECHO_REQUEST packets to network hosts.

 Options controlling ICMP request types:
      --address              send ICMP_ADDRESS packets (root only)
      --echo                 send ICMP_ECHO packets (default)
      --mask                 same as --address
      --timestamp            send ICMP_TIMESTAMP packets
  -t, --type=TYPE            send TYPE packets

 Options valid for all request types:

  -c, --count=NUMBER         stop after sending NUMBER packets
  -d, --debug                set the SO_DEBUG option
  -i, --interval=NUMBER      wait NUMBER seconds between sending each packet
  -n, --numeric              do not resolve host addresses
  -r, --ignore-routing       send directly to a host on an attached network
      --ttl=N                specify N as time-to-live
  -T, --tos=NUM              set type of service (TOS) to NUM
  -v, --verbose              verbose output
  -w, --timeout=N            stop after N seconds
  -W, --linger=N             number of seconds to wait for response

 Options valid for --echo requests:

  -f, --flood                flood ping (root only)
      --ip-timestamp=FLAG    IP timestamp of type FLAG, which is one of
                             "tsonly" and "tsaddr"
  -l, --preload=NUMBER       send NUMBER packets as fast as possible before
                             falling into normal mode of behavior (root only)
  -p, --pattern=PATTERN      fill ICMP packet with given pattern (hex)
  -q, --quiet                quiet output
  -R, --route                record route
  -s, --size=NUMBER          send NUMBER data octets

  -?, --help                 give this help list
      --usage                give a short usage message
  -V, --version              print program version
""".strip()


@ping_interface
class InetToolsPingInterface(PingInterface):

    def match_ping_usage(self, usage):
        return usage.startswith(INET_TOOLS_PING_USAGE)

    def get_deadline_option(self, parameters):
        ip_version = _parameters.get_ping_ip_version(parameters)
        if ip_version == constants.IP_VERSION_6:
            return ['-w', parameters.deadline]
        else:
            return ['-w', parameters.deadline,
                    '-W', parameters.deadline]


IPUTILS_PING_USAGE = """
ping: invalid option -- '-'
Usage: ping [-aAbBdDfhLnOqrRUvV] [-c count] [-i interval] [-I interface]
            [-m mark] [-M pmtudisc_option] [-l preload] [-p pattern] [-Q tos]
            [-s packetsize] [-S sndbuf] [-t ttl] [-T timestamp_option]
            [-w deadline] [-W timeout] [hop1 ...] destination
""".strip()


@ping_interface
class BsdPingInterface(PingInterface):

    def match_ping_usage(self, usage):
        return usage.startswith(BSD_PING_USAGE)

    has_fragment_option = False

    def get_deadline_option(self, parameters):
        return ['-t', parameters.deadline]


BSD_PING_USAGE = """
ping: unrecognized option `--help'
usage: ping [-AaDdfnoQqRrv] [-c count] [-G sweepmaxsize]
            [-g sweepminsize] [-h sweepincrsize] [-i wait]
            [-l preload] [-M mask | time] [-m ttl] [-p pattern]
            [-S src_addr] [-s packetsize] [-t timeout][-W waittime]
            [-z tos] host
       ping [-AaDdfLnoQqRrv] [-c count] [-I iface] [-i wait]
            [-l preload] [-M mask | time] [-m ttl] [-p pattern] [-S src_addr]
            [-s packetsize] [-T ttl] [-t timeout] [-W waittime]
            [-z tos] mcast-group
""".strip()



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import enum
import typing  # noqa

from oslo_log import log

import tobiko
from tobiko.shell.sh import _command
from tobiko.shell.sh import _uptime
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class RebootHostMethod(enum.Enum):

    SOFT = '/sbin/reboot',
    HARD = 'echo 1 > /proc/sys/kernel/sysrq && echo b > /proc/sysrq-trigger',
    CRASH = 'echo 1 > /proc/sys/kernel/sysrq && echo c > /proc/sysrq-trigger',

    def __init__(self, command: str):
        self.command = command


class RebootHostError(tobiko.TobikoException):
    message = "host {hostname!r} not rebooted: {cause}"


class RebootHostTimeoutError(RebootHostError):
    message = "host {hostname!r} not rebooted after {timeout!s} seconds"


def reboot_host(ssh_client: ssh.SSHClientFixture,
                wait: bool = True,
                timeout: tobiko.Seconds = None,
                method: RebootHostMethod = RebootHostMethod.SOFT):
    reboot = RebootHostOperation(ssh_client=ssh_client,
                                 timeout=timeout,
                                 method=method)
    tobiko.setup_fixture(reboot)
    if wait:
        reboot.wait_for_operation()
    return reboot


class RebootHostOperation(tobiko.Operation):

    default_wait_timeout = 300.
    default_wait_interval = 5.
    default_wait_count = 60

    def __init__(self,
                 ssh_client: ssh.SSHClientFixture,
                 timeout: tobiko.Seconds = None,
                 method: RebootHostMethod = RebootHostMethod.SOFT):
        super(RebootHostOperation, self).__init__()
        tobiko.check_valid_type(ssh_client, ssh.SSHClientFixture)
        tobiko.check_valid_type(method, RebootHostMethod)
        self.is_rebooted = False
        self.method = method
        self.ssh_client = ssh_client
        self.start_time: tobiko.Seconds = None
        self.timeout = tobiko.to_seconds(timeout)

    def run_operation(self):
        self.is_rebooted = False
        self.start_time = None
        for attempt in tobiko.retry(
                timeout=self.timeout,
                default_timeout=self.default_wait_timeout,
                default_count=self.default_wait_count,
                default_interval=self.default_wait_interval):
            try:
                channel = self.ssh_client.connect(
                    connection_timeout=attempt.time_left,
                    retry_count=1)
                LOG.info("Executing reboot command on host "
                         f"'{self.hostname}' (command='{self.command}')... ")
                self.start_time = tobiko.time()
                channel.exec_command(str(self.command))
            except Exception as ex:
                if attempt.time_left > 0.:
                    LOG.debug(f"Unable to reboot remote host "
                              f"(time_left={attempt.time_left}): {ex}")
                else:
                    LOG.exception(f"Unable to reboot remote host: {ex}")
                    raise RebootHostTimeoutError(
                        hostname=self.hostname or self.ssh_client.host,
                        timeout=attempt.timeout) from ex
            else:
                LOG.info(f"Host '{self.hostname}' is rebooting "
                         f"(command='{self.command}').")
                break
            finally:
                # Ensure we close connection after rebooting command
                self.ssh_client.close()

    def cleanup_fixture(self):
        self.is_rebooted = False
        self.start_time = None

    @property
    def command(self) -> _command.ShellCommand:
        return _command.shell_command(
            ['sudo', '/bin/sh', '-c', self.method.command])

    @property
    def hostname(self) -> str:
        return self.ssh_client.hostname

    @property
    def elapsed_time(self) -> tobiko.Seconds:
        if self.start_time is None:
            return None
        else:
            return tobiko.time() - self.start_time

    @property
    def time_left(self) -> tobiko.Seconds:
        if self.timeout is None or self.elapsed_time is None:
            return None
        else:
            return self.timeout - self.elapsed_time

    def wait_for_operation(self, timeout: tobiko.Seconds = None):
        if self.is_rebooted:
            return
        try:
            for attempt in tobiko.retry(
                    timeout=tobiko.min_seconds(timeout, self.time_left),
                    default_timeout=self.default_wait_timeout,
                    default_count=self.default_wait_count,
                    default_interval=self.default_wait_interval):
                # ensure SSH connection is closed before retrying connecting
                tobiko.cleanup_fixture(self.ssh_client)
                assert self.ssh_client.client is None
                LOG.debug(f"Getting uptime after reboot '{self.hostname}' "
                          "after reboot... ")
                try:
                    up_time = _uptime.get_uptime(ssh_client=self.ssh_client,
                                                 timeout=30.)

                except Exception:
                    # if disconnected while getting up time we assume the VM is
                    # just rebooting. These are good news!
                    LOG.debug("Unable to get uptime from host "
                              f"'{self.hostname}'", exc_info=1)
                    attempt.check_limits()
                else:
                    # verify that reboot actually happened by comparing elapsed
                    # time with up_time
                    elapsed_time = self.elapsed_time
                    if up_time < elapsed_time:
                        assert self.ssh_client.client is not None
                        self.is_rebooted = True
                        LOG.debug(f"Host '{self.hostname}' restarted "
                                  f"{elapsed_time} seconds after "
                                  f"reboot operation (up_time={up_time})")
                        break
                    else:
                        LOG.debug(f"Host '{self.hostname}' still not "
                                  f"restarted {elapsed_time} seconds after "
                                  f"reboot operation (up_time={up_time!r})")
                        attempt.check_limits()
        finally:
            if not self.is_rebooted:
                self.ssh_client.close()



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import pytest
import testtools

from tobiko.openstack import neutron
from tobiko.openstack import tests


@neutron.skip_unless_is_ovn()
class NodeTest(testtools.TestCase):

    @pytest.mark.ovn_migration
    def test_ovs_namespaces_are_absent(self):
        tests.test_ovs_namespaces_are_absent()

    @pytest.mark.ovn_migration
    def test_ovs_interfaces_are_absent(self):
        tests.test_ovs_interfaces_are_absent()



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


class NoMatchingLinesFound(tobiko.TobikoException):
    message = ("No matching lines found in files (pattern='{pattern}',"
               " files={files}, login={login})")


def grep(pattern: str,
         command: typing.Optional[sh.ShellCommandType] = None,
         grep_command: sh.ShellCommandType = 'zgrep -Eh',
         files: typing.Optional[typing.List[str]] = None,
         ssh_client: ssh.SSHClientFixture = None,
         blank_lines: bool = True,
         **execute_params) -> typing.List[str]:
    if not pattern:
        raise ValueError("Pattern string can't be empty")
    if command:
        if files:
            raise ValueError("File list must be empty when command is given")
        command_line = sh.shell_command(command) + ['|'] + grep_command + [
            '-e', pattern]
    elif files:
        command_line = sh.shell_command(grep_command) + [
            '-e', pattern] + files
    else:
        raise ValueError("command and files can't be both empty or None")

    try:
        stdout = sh.execute(command_line,
                            ssh_client=ssh_client,
                            **execute_params).stdout
    except sh.ShellCommandFailed as ex:
        if ex.exit_status > 1:
            # Some unknown problem occurred
            raise
        stdout = ex.stdout

    output_lines: typing.List[str] = [
        line
        for line in stdout.splitlines()
        if blank_lines or line.strip()]
    if output_lines:
        return output_lines

    login = ssh_client.login if ssh_client else None
    raise NoMatchingLinesFound(pattern=pattern,
                               files=files,
                               login=login)


def grep_files(pattern: str,
               files: typing.List[str],
               **grep_params) -> typing.List[str]:
    return grep(pattern=pattern, files=files, **grep_params)


def grep_lines(pattern: str,
               command: sh.ShellCommandType,
               **grep_params) -> typing.List[str]:
    return grep(pattern=pattern, command=command, **grep_params)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


GROUP_NAME = 'glance'
OPTIONS = [
    cfg.StrOpt('image_dir',
               default='~/.tobiko/cache/glance/images',
               help=("Default directory where to look for image "
                     "files")),
]


GLANCE_IMAGE_NAMES = ['centos',
                      'centos7',
                      'cirros',
                      'fedora',
                      'rhel',
                      'ubuntu']


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)

    for image_options in get_images_options():
        conf.register_opts(group=image_options[0], opts=image_options[1])


def get_images_options():
    options = []
    for name in GLANCE_IMAGE_NAMES:
        group_name = name.lower()
        options += [(
            group_name,
            [cfg.StrOpt('image_name',
                        help="Default " + name + " image name"),
             cfg.StrOpt('image_url',
                        help="Default " + name + " image URL"),
             cfg.StrOpt('image_file',
                        help="Default " + name + " image filename"),
             cfg.StrOpt('container_format',
                        help="Default " + name + " container format"),
             cfg.StrOpt('disk_format',
                        help="Default " + name + " disk format"),
             cfg.StrOpt('username',
                        help="Default " + name + " username"),
             cfg.StrOpt('password',
                        help="Default " + name + " password"),
             cfg.FloatOpt('connection_timeout',
                          default=None,
                          help=("Default " + name +
                                " SSH connection timeout (seconds)")), ]
        )]

    return options


def list_options():
    options = [(GROUP_NAME, itertools.chain(OPTIONS))]
    for image_options in get_images_options():
        options += [
            (image_options[0], itertools.chain(image_options[1]))]
    return options



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import shlex
import typing


ShellCommandType = typing.Union['ShellCommand', str, typing.Iterable[str]]


class ShellCommand(tuple):

    def __repr__(self) -> str:
        return f"ShellCommand({str(self)!r})"

    def __str__(self) -> str:
        return join(self)

    def __add__(self, other: ShellCommandType) -> 'ShellCommand':
        return shell_command(tuple(self) + shell_command(other))


def shell_command(command: ShellCommandType,
                  **shlex_params) -> ShellCommand:
    if isinstance(command, ShellCommand):
        return command
    elif isinstance(command, str):
        return split(command, **shlex_params)
    else:
        return ShellCommand(str(a) for a in command)


_find_unsafe = re.compile(r'[^\w@&%+=:,.;<>/\-()\[\]|*~]', re.ASCII).search

_is_quoted = re.compile(r'(^\'.*\'$)|(^".*"$)', re.ASCII).search


def quote(s: str):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"

    if _is_quoted(s):
        return s

    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


def join(sequence: typing.Iterable[str]) -> str:
    return ' '.join(quote(s)
                    for s in sequence)


def split(command: str, posix=True, **shlex_params) -> ShellCommand:
    lex = shlex.shlex(command, posix=posix, **shlex_params)
    lex.whitespace_split = True
    return ShellCommand(lex)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import itertools
import sys
import typing

from oslo_log import log
import testtools

from tobiko.common import _asserts
from tobiko.common import _exception
from tobiko.common import _time


LOG = log.getLogger(__name__)


class RetryException(_exception.TobikoException):
    pass


class RetryLimitError(RetryException):
    message = ("Retry limit exceeded ({attempt.details})")


class RetryCountLimitError(RetryLimitError):
    message = ("Retry count limit exceeded ({attempt.details})")


class RetryTimeLimitError(RetryLimitError):
    message = ("Retry time limit exceeded ({attempt.details})")


class RetryAttempt(object):

    def __init__(self,
                 number: int,
                 start_time: float,
                 elapsed_time: float,
                 count: typing.Optional[int] = None,
                 timeout: _time.Seconds = None,
                 interval: _time.Seconds = None):
        self.number = number
        self.start_time = start_time
        self.elapsed_time = elapsed_time
        self.count = count
        self.timeout = _time.to_seconds(timeout)
        self.interval = _time.to_seconds(interval)

    def __eq__(self, other):
        return (other.number == self.number and
                other.start_time == self.start_time and
                other.elapsed_time == self.elapsed_time and
                other.count == self.count and
                other.timeout == self.timeout and
                other.interval == self.interval)

    def __hash__(self):
        raise NotImplementedError

    @property
    def count_left(self) -> typing.Optional[int]:
        if self.count is None:
            return None
        else:
            return max(0, self.count - self.number)

    def check_count_left(self) -> _time.Seconds:
        if self.count_left == 0:
            _exception.exc_info().reraise()
            raise RetryCountLimitError(attempt=self)
        return self.count_left

    @property
    def time_left(self) -> _time.Seconds:
        if self.timeout is None:
            return None
        else:
            return max(0., self.timeout - self.elapsed_time)

    def check_time_left(self) -> _time.Seconds:
        if self.time_left == 0.:
            _exception.exc_info().reraise()
            raise RetryTimeLimitError(attempt=self)
        return self.time_left

    def check_limits(self):
        self.check_count_left()
        self.check_time_left()

    @property
    def details(self) -> str:
        details = [f"number={self.number}"]
        if self.count is not None:
            details.append(f"count={self.count}")
        details.append(f"elapsed_time={self.elapsed_time}")
        if self.timeout is not None:
            details.append(f"timeout={self.timeout}")
        if self.interval is not None:
            details.append(f"interval={self.interval}")
        return ', '.join(details)

    @property
    def is_last(self):
        try:
            self.check_limits()
        except Exception:
            return True
        else:
            return False

    def __repr__(self):
        return f"retry_attempt({self.details})"


def retry_attempt(number: int = 0,
                  start_time: float = 0.,
                  elapsed_time: float = 0.,
                  count: typing.Optional[int] = None,
                  timeout: _time.Seconds = None,
                  interval: _time.Seconds = None) -> RetryAttempt:
    return RetryAttempt(number=number,
                        count=count,
                        start_time=start_time,
                        elapsed_time=elapsed_time,
                        timeout=timeout, interval=interval)


class Retry(object):

    def __init__(self,
                 count: typing.Optional[int] = None,
                 timeout: _time.Seconds = None,
                 interval: _time.Seconds = None):
        self.count = count
        self.timeout = _time.to_seconds(timeout)
        self.interval = _time.to_seconds(interval)

    def __eq__(self, other):
        return (other.count == self.count and
                other.timeout == self.timeout and
                other.interval == self.interval)

    def __hash__(self):
        raise NotImplementedError

    def __iter__(self) -> typing.Iterator[RetryAttempt]:
        start_time = _time.time()
        elapsed_time = 0.
        for number in itertools.count(1):
            attempt = retry_attempt(number=number,
                                    count=self.count,
                                    start_time=start_time,
                                    elapsed_time=elapsed_time,
                                    timeout=self.timeout,
                                    interval=self.interval)

            yield attempt

            attempt.check_limits()

            elapsed_time = _time.time() - start_time
            interval = self.interval
            if interval is not None:
                sleep_time = attempt.number * interval - elapsed_time
                sleep_time = max(0., sleep_time)
                time_left = attempt.time_left
                if sleep_time > 0.:
                    if time_left is None or time_left > sleep_time:
                        LOG.debug(f"Wait for {sleep_time} seconds before "
                                  f"retrying... ({attempt.details})")
                        _time.sleep(sleep_time)
                        elapsed_time = _time.time() - start_time

    @property
    def details(self) -> str:
        details = []
        if self.count is not None:
            details.append(f"count={self.count}")
        if self.timeout is not None:
            details.append(f"timeout={self.timeout}")
        if self.interval is not None:
            details.append(f"interval={self.interval}")
        return ', '.join(details)

    def __repr__(self):
        return f"retry({self.details})"


def retry(other_retry: typing.Optional[Retry] = None,
          count: typing.Optional[int] = None,
          timeout: _time.Seconds = None,
          interval: _time.Seconds = None,
          default_count: typing.Optional[int] = None,
          default_timeout: _time.Seconds = None,
          default_interval: _time.Seconds = None) -> Retry:

    if other_retry is not None:
        # Apply default values from the other Retry object
        _exception.check_valid_type(other_retry, Retry)
        count = count or other_retry.count
        timeout = timeout or other_retry.timeout
        interval = interval or other_retry.interval

    # Apply default values
    count = count or default_count
    timeout = timeout or default_timeout
    interval = interval or default_interval

    return Retry(count=count, timeout=timeout, interval=interval)


def retry_on_exception(
        exception: Exception,
        *exceptions: Exception,
        other_retry: typing.Optional[Retry] = None,
        count: typing.Optional[int] = None,
        timeout: _time.Seconds = None,
        interval: _time.Seconds = None,
        default_count: typing.Optional[int] = None,
        default_timeout: _time.Seconds = None,
        default_interval: _time.Seconds = None,
        on_exception: typing.Optional[typing.Callable] = None) -> \
        typing.Callable[[typing.Callable], typing.Callable]:

    retry_object = retry(other_retry=other_retry,
                         count=count,
                         timeout=timeout,
                         interval=interval,
                         default_count=default_count,
                         default_timeout=default_timeout,
                         default_interval=default_interval)
    exceptions = (exception,) + exceptions

    def decorator(func):
        if typing.TYPE_CHECKING:
            # Don't neet to wrap the function when going to check argument
            # types
            return func

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # pylint: disable=catching-non-exception
            for attempt in retry_object:
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    attempt.check_limits()
                    if on_exception is not None:
                        on_exception(attempt, *args, **kwargs)
        return wrapper

    return decorator


def retry_test_case(*exceptions: Exception,
                    count: typing.Optional[int] = None,
                    timeout: _time.Seconds = None,
                    interval: _time.Seconds = None) -> \
                    typing.Callable[[typing.Callable], typing.Callable]:
    """Re-run test case method in case it fails
    """
    exceptions = exceptions or (_asserts.FailureException,)
    return retry_on_exception(*exceptions,
                              count=count,
                              timeout=timeout,
                              interval=interval,
                              default_count=3,
                              on_exception=on_test_case_retry_exception)


def on_test_case_retry_exception(attempt: RetryAttempt,
                                 test_case: testtools.TestCase,
                                 *_args, **_kwargs):
    # pylint: disable=protected-access
    _exception.check_valid_type(test_case, testtools.TestCase)
    test_case._report_traceback(sys.exc_info(),
                                f"traceback[attempt={attempt.number}]")
    LOG.exception("Re-run test after failed attempt. "
                  f"(attempt={attempt.number}, test='{test_case.id()}')")



from __future__ import absolute_import

import json
import re

from oslo_log import log

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.tripleo import pacemaker

LOG = log.getLogger(__name__)


def build_ovn_db_show_dict(ovn_db_show_str):
    # returns a dictionary with OVN NB or OVN SB DB information
    # each dict key is a section from the OVN DB command output
    # each dict value is a list of strings
    ovn_master_db_dict = {}
    for line in ovn_db_show_str.splitlines():
        if not re.match(r'^\s+', line):
            # if line does not start with a space char, a new section will be
            # started, i.e. a new dict entry
            current_ovn_section = line.strip()
            ovn_master_db_dict[current_ovn_section] = []
        else:
            ovn_master_db_dict[current_ovn_section].append(line.strip())

    return ovn_master_db_dict


def test_neutron_agents_are_alive(timeout=300., interval=5.) \
        -> tobiko.Selection[neutron.NeutronAgentType]:
    for attempt in tobiko.retry(timeout=timeout, interval=interval):
        LOG.debug("Look for unhealthy Neutron agents...")
        try:
            # get Neutron agent list
            agents = neutron.list_agents()
        except (neutron.ServiceUnavailable,
                neutron.NeutronClientException) as ex:
            if attempt.is_last:
                raise
            else:
                # retry because Neutron server could still be unavailable
                # after a disruption
                LOG.debug(f"Waiting for neutron service... ({ex})")
                continue  # Let retry

        dead_agents = agents.with_items(alive=False)
        if dead_agents:
            dead_agents_details = json.dumps(agents, indent=4, sort_keys=True)
            if attempt.is_last:
                tobiko.fail("Unhealthy agent(s) found:\n"
                            f"{dead_agents_details}\n")
            else:
                # retry because some Neutron agent could still be unavailable
                # after a disruption
                LOG.debug("Waiting for Neutron agents to get alive...\n"
                          f"{dead_agents_details}")
                continue

        LOG.debug(f"All {len(agents)} Neutron agents are alive.")
        break
    else:
        raise RuntimeError("Retry loop broken")

    return agents


def ovn_dbs_are_synchronized(test_case):
    from tobiko.tripleo import containers
    # declare commands
    runtime_name = containers.get_container_runtime_name()
    search_container_cmd = (
        "%s ps --format '{{.Names}}' -f name=ovn-dbs-bundle" %
        runtime_name)
    container_cmd_prefix = ('%s exec -uroot {container}' %
                            runtime_name)
    ovndb_sync_cmd = ('ovs-appctl -t /var/run/openvswitch/{ovndb_ctl_file} '
                      'ovsdb-server/sync-status')
    ovndb_show_cmd = '{ovndb} show'
    ovndb_ctl_file_dict = {'nb': 'ovnnb_db.ctl', 'sb': 'ovnsb_db.ctl'}
    ovndb_dict = {'nb': 'ovn-nbctl', 'sb': 'ovn-sbctl'}
    expected_state_active_str = 'state: active'
    expected_state_backup_str = 'state: backup'

    # use ovn master db as a reference
    ovn_master_node_name = pacemaker.get_ovn_db_master_node()
    test_case.assertEqual(1, len(ovn_master_node_name))
    ovn_master_node = topology.get_openstack_node(ovn_master_node_name[0])
    ovn_master_dbs_show_dict = {}
    # obtained the container name
    container_name = sh.execute(
        search_container_cmd,
        ssh_client=ovn_master_node.ssh_client,
        sudo=True).stdout.splitlines()[0]
    for db in ('nb', 'sb'):
        # check its synchronization is active
        sync_cmd = (' '.join((container_cmd_prefix, ovndb_sync_cmd)).
                    format(container=container_name,
                           ovndb_ctl_file=ovndb_ctl_file_dict[db]))
        sync_status = sh.execute(sync_cmd,
                                 ssh_client=ovn_master_node.ssh_client,
                                 sudo=True).stdout
        test_case.assertIn(expected_state_active_str, sync_status)
        # obtain nb and sb show output
        show_cmd = (' '.join((container_cmd_prefix, ovndb_show_cmd)).
                    format(container=container_name, ovndb=ovndb_dict[db]))
        ovn_db_show = sh.execute(
            show_cmd, ssh_client=ovn_master_node.ssh_client, sudo=True).stdout
        ovn_master_dbs_show_dict[db] = build_ovn_db_show_dict(ovn_db_show)

    # ovn dbs are located on the controller nodes
    for node in topology.list_openstack_nodes(group='controller'):
        if node.name == ovn_master_node.name:
            # master node is the reference and do not need to be checked again
            continue
        container_name = sh.execute(
            search_container_cmd,
            ssh_client=node.ssh_client, sudo=True).stdout.splitlines()[0]
        # verify ovn nb and sb dbs are synchronized
        ovn_dbs_show_dict = {}
        for db in ('nb', 'sb'):
            # check its synchronization is active
            sync_cmd = (' '.join((container_cmd_prefix, ovndb_sync_cmd)).
                        format(container=container_name,
                               ovndb_ctl_file=ovndb_ctl_file_dict[db]))
            sync_status = sh.execute(sync_cmd,
                                     ssh_client=node.ssh_client,
                                     sudo=True).stdout
            test_case.assertIn(expected_state_backup_str, sync_status)
            # obtain nb and sb show output
            show_cmd = (' '.join((container_cmd_prefix, ovndb_show_cmd)).
                        format(container=container_name, ovndb=ovndb_dict[db]))
            ovn_db_show = sh.execute(
                show_cmd, ssh_client=node.ssh_client, sudo=True).stdout
            ovn_dbs_show_dict[db] = build_ovn_db_show_dict(ovn_db_show)
            test_case.assertEqual(len(ovn_dbs_show_dict[db]),
                                  len(ovn_master_dbs_show_dict[db]))
            for key in ovn_dbs_show_dict[db]:
                test_case.assertEqual(
                    sorted(ovn_dbs_show_dict[db][key]),
                    sorted(ovn_master_dbs_show_dict[db][key]))

    LOG.info("All OVN DBs are synchronized")


def ovn_dbs_vip_bindings(test_case):
    # commands to obtain OVN SB and NB connection strings
    get_ovn_nb_conn_cmd = (
        'crudini --get /var/lib/config-data/puppet-generated/neutron/etc/'
        'neutron/plugins/ml2/ml2_conf.ini ovn ovn_nb_connection')
    get_ovn_sb_conn_cmd = get_ovn_nb_conn_cmd.replace('ovn_nb_connection',
                                                      'ovn_sb_connection')

    controllers = topology.list_openstack_nodes(group='controller')
    ovn_conn_str = {}
    ovn_conn_str['nb'] = sh.execute(get_ovn_nb_conn_cmd,
                                    ssh_client=controllers[0].ssh_client,
                                    sudo=True).stdout.splitlines()[0]
    ovn_conn_str['sb'] = sh.execute(get_ovn_sb_conn_cmd,
                                    ssh_client=controllers[0].ssh_client,
                                    sudo=True).stdout.splitlines()[0]
    ovn_conn = {}
    for db in ('nb', 'sb'):
        ovn_conn[db] = {}
        ipv6 = re.findall(r'\[.*\]', ovn_conn_str[db])
        if len(ipv6) == 1:
            ovn_conn[db]['ip'] = ipv6[0]
        elif len(ipv6) == 0:
            ovn_conn[db]['ip'] = ovn_conn_str[db].split(':')[1]
        else:
            raise RuntimeError('Error parsing ovn db connection string from '
                               'configuration file')
        ovn_conn[db]['port'] = ovn_conn_str[db].split(':')[-1]

    # ovn db sockets might be centrillized or distributed
    # that depends on the openstack version under test
    ovn_db_sockets_centrallized = topology.verify_osp_version(
        '14.0', lower=True)

    # command to obtain sockets listening on OVN SB and DB DBs
    get_ovn_db_sockets_listening_cmd = \
        "ss -p state listening 'sport = {srcport} and src {srcip}'"

    num_db_sockets = 0
    for controller in controllers:
        for db in ('nb', 'sb'):
            ovn_db_sockets_listening = sh.execute(
                get_ovn_db_sockets_listening_cmd.format(
                    srcport=ovn_conn[db]['port'],
                    srcip=ovn_conn[db]['ip']),
                ssh_client=controller.ssh_client,
                sudo=True).stdout.splitlines()
            if ovn_db_sockets_centrallized:
                if 2 == len(ovn_db_sockets_listening):
                    num_db_sockets += 1
                    test_case.assertIn('ovsdb-server',
                                       ovn_db_sockets_listening[1])
            else:
                num_db_sockets += 1
                test_case.assertEqual(2, len(ovn_db_sockets_listening))
                test_case.assertIn('ovsdb-server', ovn_db_sockets_listening[1])

    if ovn_db_sockets_centrallized:
        test_case.assertEqual(2, num_db_sockets)
    else:
        test_case.assertEqual(2 * len(controllers), num_db_sockets)


def test_ovn_dbs_validations():
    if not neutron.has_ovn():
        LOG.debug('OVN not configured. OVN DB sync validations skipped')
        return

    test_case = tobiko.get_test_case()

    # run validations
    ovn_dbs_are_synchronized(test_case)
    ovn_dbs_vip_bindings(test_case)


def test_ovs_bridges_mac_table_size():
    test_case = tobiko.get_test_case()
    expected_mac_table_size = '50000'
    get_mac_table_size_cmd = ('ovs-vsctl get bridge {br_name} '
                              'other-config:mac-table-size')
    if neutron.has_ovn():
        get_br_mappings_cmd = ('ovs-vsctl get Open_vSwitch . '
                               'external_ids:ovn-bridge-mappings')
    else:
        get_br_mappings_cmd = (
            'crudini --get /var/lib/config-data/puppet-generated/neutron/'
            'etc/neutron/plugins/ml2/openvswitch_agent.ini '
            'ovs bridge_mappings')
    for node in topology.list_openstack_nodes(group='overcloud'):
        try:
            br_mappings_str = sh.execute(get_br_mappings_cmd,
                                         ssh_client=node.ssh_client,
                                         sudo=True).stdout.splitlines()[0]
        except sh.ShellCommandFailed:
            LOG.debug(f"bridge mappings not configured on node '{node.name}'",
                      exc_info=1)
            continue
        br_list = [br_mapping.split(':')[1] for br_mapping in
                   br_mappings_str.replace('"', '').split(',')]
        for br_name in br_list:
            mac_table_size = sh.execute(
                get_mac_table_size_cmd.format(br_name=br_name),
                ssh_client=node.ssh_client, sudo=True).stdout.splitlines()[0]
            test_case.assertEqual(mac_table_size.replace('"', ''),
                                  expected_mac_table_size)


def test_ovs_namespaces_are_absent():
    ovs_specific_namespaces = ['qrouter', 'qdhcp', 'snat', 'fip']
    for node in topology.list_openstack_nodes():
        if node.name.startswith('undercloud'):
            continue
        namespaces = ip.list_network_namespaces(
            ssh_client=node.ssh_client, sudo=True)
        namespaces = [namespace
                      for namespace in namespaces
                      if any(namespace.startswith(prefix)
                             for prefix in ovs_specific_namespaces)]
        test_case = tobiko.get_test_case()
        test_case.assertEqual(
            [], namespaces,
            f"Unexpected namespace found on {node.name}: {*namespaces,}")


def test_ovs_interfaces_are_absent():
    ovs_specific_interfaces = ['qvo', 'qvb', 'qbr', 'qr', 'qg', 'fg', 'sg']
    for node in topology.list_openstack_nodes():
        if node.name.startswith('undercloud'):
            continue
        interfaces = ip.list_network_interfaces(
            ssh_client=node.ssh_client, sudo=True)
        interfaces = [interface
                      for interface in interfaces
                      if any(interface.startswith(prefix)
                             for prefix in ovs_specific_interfaces)]
        test_case = tobiko.get_test_case()
        test_case.assertEqual(
            [], interfaces,
            f"Unexpected interface found on {node.name}: {*interfaces,}")



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
from urllib import parse
import typing

from oslo_log import log

from tobiko.docker import _exception
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


def discover_docker_urls(
        ssh_client: ssh.SSHClientType = None,
        default_url='unix:/var/run/docker.sock',
        check_urls=True,
        sudo=False):
    urls = []
    processes = sh.list_processes(command='^dockerd',
                                  ssh_client=ssh_client,
                                  sudo=sudo)

    for process in processes:
        if process.command_line:
            urls += urls_from_command_line(process.command_line)

    urls.append(default_url)
    urls = list(collections.OrderedDict.fromkeys(urls))

    error: typing.Optional[Exception] = None
    if check_urls:
        valid_urls = []
        for url in urls:
            parsed_url = parse.urlparse(url)
            if (parsed_url.scheme == 'unix' and
                    parsed_url.path.startswith('/')):
                try:
                    sh.execute(f"test -r '{parsed_url.path}'",
                               ssh_client=ssh_client,
                               sudo=sudo)
                except sh.ShellCommandFailed as ex:
                    LOG.exception(
                        f"Can't read from socket: {parsed_url.path}")
                    ex.__cause__ = error
                    error = ex
                else:
                    valid_urls.append(url)

        if not valid_urls:
            raise _exception.DockerUrlNotFoundError(
                "Docker is not running") from error
        urls = valid_urls
    return urls


def urls_from_command_line(command_line: sh.ShellCommand,
                           default_url='unix:/var/run/docker.sock') \
        -> typing.List[str]:
    urls = []
    arg_is_url = False
    for arg in command_line[1:]:
        if arg == '-H':
            arg_is_url = True
            continue
        if arg_is_url:
            arg_is_url = False
            try:
                url = parse.urlparse(arg)
            except Exception:
                LOG.debug(f'Invalid URL: {arg}', exc_info=1)
                continue
            if url.scheme == 'fd':
                urls.append(default_url)
                continue
            if url.scheme != 'unix':
                LOG.debug(f'Unsupported URL scheme: {arg}')
                continue
            if not url.path.startswith('/'):
                LOG.debug(f'Unsupported URL path: {arg}')
                continue
            urls.append(arg)
    return urls


def is_docker_running(ssh_client=None, sudo=False, **execute_params):
    try:
        discover_docker_urls(ssh_client=ssh_client,
                             sudo=sudo,
                             **execute_params)
    except _exception.DockerUrlNotFoundError:
        return False
    else:
        return True



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from docker import client as docker_client
from docker.models import containers
import testtools

import tobiko
from tobiko import docker
from tobiko.openstack import topology
from tobiko.shell import ssh


class DockerSSHClientFixture(tobiko.SharedFixture):

    ssh_client: ssh.SSHClientFixture

    def setup_fixture(self):
        self.ssh_client = self.get_ssh_client()

    def get_ssh_client(self) -> ssh.SSHClientFixture:
        for ssh_client in self.iter_ssh_clients():
            if docker.is_docker_running(ssh_client=ssh_client,
                                        sudo=True):
                return ssh_client
        tobiko.skip_test('Docker is not running')

    @staticmethod
    def iter_ssh_clients():
        ssh_client = ssh.ssh_proxy_client()
        if isinstance(ssh_client, ssh.SSHClientFixture):
            yield ssh_client

        nodes = topology.list_openstack_nodes()
        for node in nodes:
            if isinstance(node.ssh_client, ssh.SSHClientFixture):
                yield node.ssh_client


class LocalDockerClientTest(testtools.TestCase):

    sudo = False

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        if docker.is_docker_running(ssh_client=False,
                                    sudo=self.sudo):
            return False
        tobiko.skip_test('Docker is not running')

    def test_get_docker_client(self):
        client = docker.get_docker_client(ssh_client=self.ssh_client,
                                          sudo=self.sudo)
        self.assertIsInstance(client, docker.DockerClientFixture)

    def test_connect_docker_client(self):
        client = docker.get_docker_client(ssh_client=self.ssh_client,
                                          sudo=self.sudo).connect()
        self.assertIsInstance(client, docker_client.DockerClient)
        client.ping()

    def test_list_docker_containers(self):
        client = docker.get_docker_client(ssh_client=self.ssh_client,
                                          sudo=self.sudo)
        for container in docker.list_docker_containers(
                client=client):
            self.assertIsInstance(container, containers.Container)


class SSHDockerClientTest(LocalDockerClientTest):

    sudo = True

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        return tobiko.setup_fixture(DockerSSHClientFixture).ssh_client



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections

import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tripleo import undercloud


@neutron.skip_if_missing_networking_agents(neutron.OPENVSWITCH_AGENT)
@neutron.skip_if_missing_networking_agents(neutron.L3_AGENT)
class OpenvswitchTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def setUp(self):
        super(OpenvswitchTest, self).setUp()
        self.ovs_agents = neutron.list_agents(agent_type="Open vSwitch agent")
        self.router_id = self.stack.network_stack.gateway_id

        self.deleted_bridges = collections.defaultdict(set)

    def tearDown(self):
        super(OpenvswitchTest, self).tearDown()
        # Try to create all bridges which were deleted during the tests
        self._create_bridges()

    def _create_bridges(self):
        for host, bridges in self.deleted_bridges.items():
            self._create_bridge(host, bridges)

    def _create_bridge(self, hostname, bridges):
        for br_name in bridges:
            agent_host = topology.get_openstack_node(hostname=hostname)
            sh.execute(
                "sudo ovs-vsctl --may-exist add-br %s" % br_name,
                ssh_client=agent_host.ssh_client)

    def _delete_bridges(self, hostname, bridges):
        for br_name in bridges:
            agent_host = topology.get_openstack_node(hostname=hostname)
            sh.execute(
                "sudo ovs-vsctl del-br %s" % br_name,
                ssh_client=agent_host.ssh_client)
            self.deleted_bridges[hostname].add(br_name)

    def _get_agent_from_host(self, hostname):
        host_shortname = tobiko.get_short_hostname(hostname)
        for agent in self.ovs_agents:
            if host_shortname == tobiko.get_short_hostname(agent['host']):
                return agent
        raise neutron.AgentNotFoundOnHost(agent_type="neutron-ovs-agent",
                                          host=hostname)

    @undercloud.skip_if_missing_undercloud
    def test_recreate_physical_bridge(self):
        # Check if vm is reachable before test
        ip_add = self.stack.ip_address
        ping.ping_until_received(ip_add).assert_replied()

        network_l3_agents = neutron.list_l3_agent_hosting_routers(
            self.router_id)
        for agent in network_l3_agents:
            # Get neutron-ovs-agent bridge mappings
            ovs_agent = self._get_agent_from_host(agent['host'])
            self._delete_bridges(
                agent['host'],
                ovs_agent['configurations']['bridge_mappings'].values())

        ping.ping_until_unreceived(ip_add).assert_not_replied()
        self._create_bridges()
        ping.ping_until_received(ip_add).assert_replied()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from neutronclient.v2_0 import client as neutronclient

from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack import test_client


class NeutronClientFixtureTest(test_client.OpenstackClientFixtureTest):

    def create_client(self, session=None):
        return neutron.NeutronClientFixture(session=session)


class GetNeutronClientTest(openstack.OpenstackTest):

    def test_get_neutron_client(self, session=None, shared=True):
        client1 = neutron.get_neutron_client(session=session, shared=shared)
        client2 = neutron.get_neutron_client(session=session, shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)
        self.assertIsInstance(client1, neutronclient.Client)
        self.assertIsInstance(client2, neutronclient.Client)

    def test_get_neutron_client_with_not_shared(self):
        self.test_get_neutron_client(shared=False)

    def test_get_neutron_client_with_session(self):
        session = keystone.get_keystone_session()
        self.test_get_neutron_client(session=session)


class NeutronClientTest(openstack.OpenstackTest):

    def test_neutron_client_with_none(self):
        default_client = neutron.get_neutron_client()
        client = neutron.neutron_client(None)
        self.assertIsInstance(client, neutronclient.Client)
        self.assertIs(default_client, client)

    def test_neutron_client_with_client(self):
        default_client = neutron.get_neutron_client()
        client = neutron.neutron_client(default_client)
        self.assertIsInstance(client, neutronclient.Client)
        self.assertIs(default_client, client)

    def test_neutron_client_with_fixture(self):
        fixture = neutron.NeutronClientFixture()
        client = neutron.neutron_client(fixture)
        self.assertIsInstance(client, neutronclient.Client)
        self.assertIs(client, fixture.client)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import sys

import testtools

import tobiko
from tobiko.tests import unit


class SomeException(tobiko.TobikoException):
    message = "message formatted with {b} and {a}"


class TestException(unit.TobikoUnitTest):

    def test_init(self, message=None, **properties):
        ex = SomeException(message, **properties)
        expected_str = message or SomeException.message
        if properties:
            expected_str = expected_str.format(**properties)
        self.assertEqual(expected_str, str(ex))
        for k, v in properties.items():
            self.assertEqual(v, getattr(ex, k))

    def test_init_with_properties(self):
        self.test_init(a='a', b='b')

    def test_init_with_message(self):
        self.test_init('{other} message', other='another')

    def test_raise(self):
        def _raise_my_exception(**kwargs):
            raise SomeException(**kwargs)
        ex = self.assertRaises(SomeException, _raise_my_exception, b=1, a=2)
        self.assertEqual('message formatted with 1 and 2', str(ex))

    def test_repr(self):
        ex = SomeException('some reason')
        self.assertEqual("SomeException('some reason')", repr(ex))

    def test_get_invalid_property(self):
        ex = SomeException(a='1', b='2')

        def _get_invalid_property():
            return ex.invalid_attribute_name

        ex = self.assertRaises(AttributeError, _get_invalid_property)
        self.assertEqual(
            "SomeException('message formatted with 2 and 1') object has no "
            "attribute 'invalid_attribute_name'", str(ex))

    def test_docstring_example(self):

        class MyException(tobiko.TobikoException):
            message = "This exception occurred because of {reason}"

        try:
            raise MyException(reason="something went wrong")
        except MyException as ex:
            self.assertEqual(
                'This exception occurred because of something went wrong',
                str(ex))
            self.assertEqual('something went wrong', ex.reason)


class TestCheckValidType(unit.TobikoUnitTest):

    def test_check_valid_type_with_no_type(self):
        self._test_check_is_valid_type_when_invalid(object())

    def test_check_valid_type_with_one_type(self):
        self._test_check_is_valid_type_when_valid(object(), object)

    def test_check_valid_type_with_two_types(self):
        self._test_check_is_valid_type_when_valid(object(), bool, object)

    def test_check_valid_type_with_one_wrong_type(self):
        self._test_check_is_valid_type_when_invalid(object(), bool)

    def test_check_valid_type_with_two_wrong_types(self):
        self._test_check_is_valid_type_when_invalid(object(), bool, int)

    def _test_check_is_valid_type_when_valid(self, obj, *valid_types):
        result = tobiko.check_valid_type(obj, *valid_types)
        self.assertIs(obj, result)

    def _test_check_is_valid_type_when_invalid(self, obj, *valid_types):
        ex = self.assertRaises(TypeError, tobiko.check_valid_type, obj,
                               *valid_types)
        tyes_str = ', '.join([str(t) for t in valid_types])
        message = ("Object {!s} is not of a valid type ({!s})"
                   ).format(repr(obj), tyes_str)

        self.assertEqual(message, str(ex))


class TestExcInfo(unit.TobikoUnitTest):

    def test_exc_info(self):
        try:
            raise RuntimeError('some error')
        except RuntimeError:
            exc_info = tobiko.exc_info()
            exc_type, exc_value, traceback = sys.exc_info()

        self.assertEqual((exc_type, exc_value, traceback), exc_info)
        self.assertIs(RuntimeError, exc_info.type)
        self.assertIs(exc_value, exc_info.value)
        self.assertIs(traceback, exc_info.traceback)

        reraised = self.assertRaises(RuntimeError, exc_info.reraise)
        self.assertIs(exc_value, reraised)


class TestListExcInfo(unit.TobikoUnitTest):

    def test_list_exc_info(self):
        result = tobiko.list_exc_infos()
        self.assertEqual([], result)

    def test_list_exc_info_with_info(self):
        error = make_exception(RuntimeError, 'error')
        result = tobiko.list_exc_infos(exc_info=error)
        self.assertEqual([error], result)

    def test_list_exc_info_handling_exception(self):
        try:
            raise RuntimeError('error')
        except RuntimeError:
            result = tobiko.list_exc_infos()
            self.assertEqual([sys.exc_info()], result)

    def test_list_exc_info_handling_empty_multiple_exceptions(self):
        error = make_exception(testtools.MultipleExceptions)
        result = tobiko.list_exc_infos(exc_info=error)
        self.assertEqual([], result)

    def test_list_exc_info_handling_multiple_exceptions(self):
        a = make_exception(RuntimeError, 'a')
        b = make_exception(ValueError, 'b')
        c = make_exception(TypeError, 'c')
        multi = make_exception(testtools.MultipleExceptions, a, b, c)
        result = tobiko.list_exc_infos(exc_info=multi)
        self.assertEqual([a, b, c], result)

    def test_list_exc_info_handling_nested_multiple_exceptions(self):
        a = make_exception(RuntimeError, 'a')
        b = make_exception(ValueError, 'b')
        c = make_exception(TypeError, 'c')
        d = make_exception(IndexError, 'd')
        inner = make_exception(testtools.MultipleExceptions, b, c)
        multi = make_exception(testtools.MultipleExceptions, a, inner, d)
        result = tobiko.list_exc_infos(exc_info=multi)
        self.assertEqual([a, b, c, d], result)


class TestHandleMultipleExceptions(unit.TobikoUnitTest):

    def test_handle_multiple_exceptions(self):
        with tobiko.handle_multiple_exceptions():
            pass

    def test_handle_multiple_exceptions_with_exception(self):
        def run():
            with tobiko.handle_multiple_exceptions():
                raise RuntimeError('error')
        self.assertRaises(RuntimeError, run)

    def test_handle_multiple_exceptions_with_empty_multiple_exception(self):
        with tobiko.handle_multiple_exceptions():
            raise testtools.MultipleExceptions()

    def test_handle_multiple_exceptions_with_multiple_exceptions(self):
        a = make_exception(TypeError, 'a')
        b = make_exception(ValueError, 'b')
        c = make_exception(RuntimeError, 'c')

        def run():
            with tobiko.handle_multiple_exceptions():
                raise testtools.MultipleExceptions(a, b, c)

        ex = self.assertRaises(TypeError, run)
        self.assertEqual(a[1], ex)

    def test_handle_multiple_exceptions_with_nested_multiple_exceptions(self):
        a = make_exception(RuntimeError, 'a')
        b = make_exception(ValueError, 'b')
        c = make_exception(TypeError, 'c')
        d = make_exception(IndexError, 'd')
        inner = make_exception(testtools.MultipleExceptions, b, c)

        def run():
            with tobiko.handle_multiple_exceptions():
                raise testtools.MultipleExceptions(a, inner, d)

        ex = self.assertRaises(RuntimeError, run)
        self.assertEqual(a[1], ex)

    def test_handle_multiple_exceptions_with_handle_exceptions(self):
        a = make_exception(RuntimeError, 'a')
        b = make_exception(ValueError, 'b')
        c = make_exception(TypeError, 'c')
        d = make_exception(IndexError, 'd')
        inner = make_exception(testtools.MultipleExceptions, b, c)

        handled_exceptions = []

        def handle_exception(ex_type, ex_value, ex_tb):
            self.assertIsInstance(ex_value, ex_type)
            handled_exceptions.append((ex_type, ex_value, ex_tb))

        def run():
            with tobiko.handle_multiple_exceptions(
                    handle_exception=handle_exception):
                raise testtools.MultipleExceptions(a, inner, d)

        ex = self.assertRaises(RuntimeError, run)
        self.assertIs(a[1], ex)
        self.assertEqual([b, c, d], handled_exceptions)


def make_exception(cls, *args, **kwargs):
    try:
        raise cls(*args, **kwargs)
    except cls:
        return sys.exc_info()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from keystoneclient import base
from keystoneclient import client as keystoneclient
from keystoneclient.v2_0 import client as v2_client
from keystoneclient.v3 import client as v3_client
from keystoneclient.v3 import endpoints as v3_endpoints

import tobiko
from tobiko.openstack import _client


class KeystoneClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return keystoneclient.Client(session=session)


class KeystoneClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return KeystoneClientFixture(session=session)


CLIENTS = KeystoneClientManager()
CLIENT_CLASSES = (v2_client.Client, v3_client.Client)
KeystoneClient = typing.Union[v2_client.Client, v3_client.Client]
KeystoneClientType = typing.Union[KeystoneClient,
                                  KeystoneClientFixture,
                                  typing.Type[KeystoneClientFixture],
                                  None]


def keystone_client(obj: KeystoneClientType) -> KeystoneClient:
    if obj is None:
        return get_keystone_client()

    if isinstance(obj, CLIENT_CLASSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, KeystoneClientFixture):
        return tobiko.setup_fixture(obj).client

    raise TypeError(f"Object {obj} is not a KeystoneClientFixture")


def get_keystone_client(session=None, shared=True, init_client=None,
                        manager=None) -> KeystoneClient:
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


_RAISE_ERROR = object()


def find_endpoint(client=None, unique=False, default=_RAISE_ERROR,
                  **attributes):
    endpoints = list_endpoints(client=client, **attributes)
    if default is _RAISE_ERROR or endpoints:
        if unique:
            return endpoints.unique
        else:
            return endpoints.first
    else:
        return default


def find_service(client=None, unique=False, default=_RAISE_ERROR, **attribute):
    services = list_services(client=client, **attribute)
    if default is _RAISE_ERROR or services:
        if unique:
            return services.unique
        else:
            return services.first
    else:
        return default


def list_endpoints(client=None, service=None, interface=None, region=None,
                   translate=True, **attributes):
    client = keystone_client(client)

    service = service or attributes.pop('service_id', None)
    if service:
        attributes['service_id'] = base.getid(service)

    region = region or attributes.pop('region_id', None)
    if region:
        attributes['region_id'] = base.getid(region)

    if client.version == 'v2.0':
        endpoints = client.endpoints.list()
        if translate:
            endpoints = translate_v2_endpoints(v2_endpoints=endpoints,
                                               interface=interface)
    else:
        endpoints = client.endpoints.list(service=service,
                                          interface=interface,
                                          region=region)
    endpoints = tobiko.select(endpoints)
    if attributes:
        endpoints = endpoints.with_attributes(**attributes)
    return endpoints


def list_services(client=None, name=None, service_type=None, **attributes):
    client = keystone_client(client)

    service_type = service_type or attributes.pop('type', None)
    if service_type:
        attributes['type'] = base.getid(service_type)

    if name:
        attributes['name'] = name

    if client.version == 'v2.0':
        services = client.services.list()
    else:
        services = client.services.list(name=name,
                                        service_type=service_type)
    services = tobiko.select(services)
    if attributes:
        services = services.with_attributes(**attributes)
    return services


def translate_v2_endpoints(v2_endpoints, interface=None):
    interfaces = interface and [interface] or v3_endpoints.VALID_INTERFACES
    endpoints = []
    for endpoint in v2_endpoints:
        for interface in interfaces:
            url = getattr(endpoint, interface + 'url')
            info = dict(id=endpoint.id,
                        interface=interface,
                        region_id=endpoint.region,
                        service_id=endpoint.service_id,
                        url=url,
                        enabled=endpoint.enabled)
            endpoints.append(v3_endpoints.Endpoint(manager=None,
                                                   info=info))
    return endpoints


def find_service_endpoint(enabled=True, interface='public', client=None,
                          **params):
    client = keystone_client(client)
    service = find_service(client=client, enabled=enabled, **params)
    return find_endpoint(client=client, service=service, interface=interface,
                         enabled=enabled)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import subprocess
import sys

TOP_DIR = os.path.dirname(os.path.dirname(__file__))
if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa

LOG = common.get_logger(__name__)


def main():
    common.setup_logging()
    show_infrared_version()

    plugin_path = os.environ.get('IR_TOBIKO_PLUGIN')
    if plugin_path:
        add_tobiko_plugin(path=plugin_path)

    show_infrared_workspaces()
    ensure_workspace()
    copy_inventory()


def show_infrared_version():
    return common.execute('ir --version || ir --version', capture_stdout=False)


def show_infrared_plugins():
    return common.execute('ir plugin list', capture_stdout=False)


def has_plugin(name):
    try:
        common.execute("ir plugin list | awk '( $4 == \"{}\" )'", name,
                       capture_stdout=False)
    except subprocess.CalledProcessError as ex:
        LOG.debug("tobiko plugin not found ({})", ex)
        return False
    else:
        LOG.info("tobiko plugin found")
        return True


def show_infrared_workspaces():
    return common.execute('ir workspace list', capture_stdout=False)


def add_tobiko_plugin(path=None):
    if has_plugin('tobiko'):
        remove_plugin('tobiko')
    add_plugin(path)
    show_infrared_plugins()


def remove_plugin(name):
    try:
        common.execute('ir plugin remove "{}"', name)
    except subprocess.CalledProcessError as ex:
        LOG.debug("plug-in '%s' not removed: %s", name, ex)
        return False
    else:
        LOG.info("plug-in '%s' removed", name)
        return True


def add_plugin(path):
    path = common.normalize_path(path)
    if not os.path.isdir(path):
        message = ("invalid plug-in directory: '{}'").format(path)
        raise RuntimeError(message)

    common.execute('ir plugin add "{}"', path)
    LOG.info("plug-in added from path '%s'", path)


def ensure_workspace(filename=None):
    filename = (filename or
                os.environ.get('IR_WORKSPACE_FILE') or
                'workspace.tgz')
    filename = common.normalize_path(filename)
    workspace = common.name_from_path(filename)
    if os.path.isfile(filename):
        try:
            common.execute('ir workspace delete "{}"', workspace)
        except subprocess.CalledProcessError as ex:
            LOG.debug("workspace '%s' not deleted: %s", workspace, ex)
        common.execute('ir workspace import "{}"', filename)
        LOG.info("workspace imported from file '%s'", filename)
        return
    else:
        LOG.debug("workspace file not found: '%s'", filename)

    try:
        common.execute('ir workspace checkout "{}"', workspace)
    except subprocess.CalledProcessError as ex:
        LOG.debug("workspace '%s' not checked out: %s", workspace, ex)
    else:
        LOG.info("workspace '%s' checked out", workspace)
        return

    common.execute('infrared workspace checkout --create "{}"', workspace)
    LOG.info("workspace '%s' created", workspace)


def copy_inventory(filename=None):
    filename = (filename or
                os.environ.get('ANSIBLE_INVENTORY') or
                'ansible_hosts')
    if not os.path.isfile(filename):
        LOG.debug('inventary file not found: %r', filename)
        return False

    dest_file = common.execute('ir workspace inventory')
    LOG.debug("got workspace inventory file: '%s'", dest_file)

    dest_dir = os.path.basename(dest_file)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
        LOG.info("directory created: '%s'", dest_dir)

    common.execute('cp {} {}', filename, dest_file)
    LOG.info("inventary file '%s' copied to '%s'", filename, dest_file)
    return True


if __name__ == '__main__':
    main()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import asyncio
import typing


class ActorRequest(typing.NamedTuple):
    future: asyncio.Future
    method: str
    arguments: typing.Dict[str, typing.Any]


class ActorRequestQueue(object):

    def __init__(self, loop: asyncio.AbstractEventLoop, max_size=0):
        self._queue: asyncio.Queue = asyncio.Queue(maxsize=max_size)
        self._loop = loop

    def send_request(self,
                     method: str,
                     arguments: typing.Dict[str, typing.Any]):
        future = self._loop.create_future()
        request = ActorRequest(future=future,
                               method=method,
                               arguments=arguments)
        self._queue.put_nowait(request)
        return future

    async def receive_request(self):
        return await self._queue.get()


def create_request_queue(loop: asyncio.AbstractEventLoop, max_size=0) \
        -> ActorRequestQueue:
    return ActorRequestQueue(loop=loop, max_size=max_size)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import mock

import tobiko
from tobiko.tests import unit
from tobiko.tests import conftest


LOG = log.getLogger(__name__)


class CaplogTest(unit.TobikoUnitTest):

    def test_configure_caplog(self):
        pytest_config = mock.MagicMock(inicfg={
            'log_level': '<existing>',
            'log_format': '<existing>',
            'log_date_format': '<existing>'})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('<existing>', pytest_config.inicfg['log_level'])
        self.assertEqual('<existing>', pytest_config.inicfg['log_format'])
        self.assertEqual('<existing>', pytest_config.inicfg['log_date_format'])

    def test_configure_caplog_debug(self):
        self.patch_caplog_config(capture_log=True, debug=True)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('DEBUG', pytest_config.inicfg['log_level'])

    def test_configure_caplog_info(self):
        self.patch_caplog_config(capture_log=True, debug=False)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('INFO', pytest_config.inicfg['log_level'])

    def test_configure_caplog_fatal(self):
        self.patch_caplog_config(capture_log=False)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('FATAL', pytest_config.inicfg['log_level'])

    def test_configure_caplog_log_format(self):
        self.patch_caplog_config(log_format='<some-format>')
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('<some-format>', pytest_config.inicfg['log_format'])

    def test_configure_caplog_log_date_format(self):
        self.patch_caplog_config(log_date_format='<some-format>')
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_caplog(pytest_config)
        self.assertEqual('<some-format>',
                         pytest_config.inicfg['log_date_format'])

    def patch_caplog_config(self, capture_log=False, debug=False,
                            log_format=None, log_date_format=None):
        tobiko_config = self.patch(tobiko, 'tobiko_config').return_value
        tobiko_config.logging.capture_log = capture_log
        tobiko_config.debug = debug
        tobiko_config.logging_default_format_string = log_format
        tobiko_config.log_date_format = log_date_format


class TimeoutTest(unit.TobikoUnitTest):

    def test_configure_timeout_existing(self):
        pytest_config = mock.MagicMock(inicfg={'timeout': '<existing>'})
        conftest.configure_timeout(pytest_config)
        self.assertEqual('<existing>', pytest_config.inicfg['timeout'])

    def test_configure_timeout_none(self):
        self.patch_timeout_config(timeout=None)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_timeout(pytest_config)
        self.assertNotIn('timeout', pytest_config.inicfg)

    def test_configure_timeout_zero(self):
        self.patch_timeout_config(timeout=0.)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_timeout(pytest_config)
        self.assertNotIn('timeout', pytest_config.inicfg)

    def test_configure_timeout_negative(self):
        self.patch_timeout_config(timeout=-1.)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_timeout(pytest_config)
        self.assertNotIn('timeout', pytest_config.inicfg)

    def test_configure_timeout_positive(self):
        self.patch_timeout_config(timeout=10.)
        pytest_config = mock.MagicMock(inicfg={})
        conftest.configure_timeout(pytest_config)
        self.assertEqual(10., pytest_config.inicfg['timeout'])

    def patch_timeout_config(self, timeout):
        tobiko_config = self.patch(tobiko, 'tobiko_config').return_value
        tobiko_config.testcase.timeout = timeout



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from urllib3 import connection
from urllib3 import connectionpool

from tobiko.shell import ssh


class HTTPConnection(connection.HTTPConnection):

    def __init__(self, *args, **kwargs):
        #: Port forwarding address to redirect connection too if given
        self.forward_address = kwargs.pop("forward_address", None)
        super(HTTPConnection, self).__init__(*args, **kwargs)

    def _new_conn(self):
        """ Establish a socket connection and set nodelay settings on it.

        :return: New socket connection.
        """
        extra_kw = {}
        if self.source_address:
            extra_kw["source_address"] = self.source_address

        if self.socket_options:
            extra_kw["socket_options"] = self.socket_options

        address = self.forward_address or (self._dns_host, self.port)
        try:
            conn = connection.connection.create_connection(
                address, self.timeout, **extra_kw)

        except connection.SocketTimeout as ex:
            raise connection.ConnectTimeoutError(
                self, (f"Connection to {self.host} timed out. "
                       f"(connect timeout={self.timeout})")) from ex

        except connection.SocketError as ex:
            raise connection.NewConnectionError(
                self, f"Failed to establish a new connection: {ex}") from ex

        return conn


class HTTPSConnection(HTTPConnection, connection.HTTPSConnection):
    pass


class HTTPConnectionPool(connectionpool.HTTPConnectionPool):

    ConnectionCls = HTTPConnection

    def __init__(self, host, port, ssh_client=None, **kwargs):
        forward_address = ssh.get_forward_port_address(address=(host, port),
                                                       ssh_client=ssh_client)
        super(HTTPConnectionPool, self).__init__(
            host=host, port=port, forward_address=forward_address, **kwargs)


class HTTPSConnectionPool(HTTPConnectionPool,
                          connectionpool.HTTPSConnectionPool):

    ConnectionCls = HTTPSConnection



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing

from tobiko import _exception


T = typing.TypeVar('T')


class Selection(list, typing.Generic[T]):

    def with_attributes(self, **attributes) -> 'Selection[T]':
        return self.select(lambda obj: equal_attributes(obj, attributes))

    def without_attributes(self, **attributes) -> 'Selection[T]':
        return self.select(lambda obj: equal_attributes(obj, attributes,
                                                        inverse=True))

    def with_items(self: 'Selection[typing.Dict]', **items) \
            -> 'Selection[typing.Dict]':
        return self.select(lambda obj: equal_items(obj, items))

    def without_items(self: 'Selection[typing.Dict]', **items) -> \
            'Selection[typing.Dict]':
        return self.select(lambda obj: equal_items(obj, items, inverse=True))

    @classmethod
    def create(cls, objects: typing.Iterable[T]) -> 'Selection[T]':
        return cls(objects)

    def select(self,
               predicate: typing.Callable[[T], bool],
               expect=True) \
            -> 'Selection[T]':
        return self.create(obj
                           for obj in self
                           if bool(predicate(obj)) is expect)

    def unselect(self,
                 predicate: typing.Callable[[T], typing.Any]) \
            -> 'Selection[T]':
        return self.select(predicate, expect=False)

    @property
    def first(self) -> T:
        if self:
            return self[0]
        else:
            raise ObjectNotFound()

    @property
    def last(self) -> T:
        if self:
            return self[-1]
        else:
            raise ObjectNotFound()

    @property
    def unique(self) -> T:
        if len(self) > 1:
            raise MultipleObjectsFound(list(self))
        else:
            return self.first

    def __repr__(self):
        return f'{type(self).__name__}({list(self)!r})'


def select(objects: typing.Iterable[T]) -> Selection[T]:
    return Selection.create(objects)


def equal_attributes(obj,
                     attributes: typing.Dict[str, typing.Any],
                     inverse=False) \
        -> bool:
    for key, matcher in attributes.items():
        matching = match(matcher, getattr(obj, key))
        if matching is inverse:
            return False
    return True


def equal_items(obj: typing.Dict,
                items: typing.Dict,
                inverse=False) -> bool:
    for key, matcher in items.items():
        matching = match(matcher, obj[key])
        if matching is inverse:
            return False
    return True


PatternType = type(re.compile("", 0))


def match(matcher: typing.Any, value: typing.Any) -> bool:
    if isinstance(matcher, PatternType):
        return matcher.match(value) is not None
    else:
        return matcher == value


class ObjectNotFound(_exception.TobikoException):
    message = "Object not found"


class MultipleObjectsFound(_exception.TobikoException):
    message = "Multiple objects found: {objects!r}"



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr

import tobiko
from tobiko.shell import sh


class IfconfigError(tobiko.TobikoException):
    message = "Unable to get IP addresses from host: {error!s}"


INETS = {
    4: ['inet'],
    6: ['inet6'],
    None: ['inet', 'inet6']
}


def list_ip_addresses(ip_version=None, **execute_params):
    inets = INETS.get(ip_version)
    if inets is None:
        error = "invalid IP version: {!r}".format(ip_version)
        raise IfconfigError(error=error)

    output = execute_ifconfig(**execute_params)

    ips = tobiko.Selection()
    for line in output.splitlines():
        if line.startswith(' '):
            try:
                fields = line.strip().split()
                if fields[0] in inets:
                    address = fields[1]
                    if address.startswith('addr:'):
                        address = address[len('addr:'):]
                        if not address:
                            address = fields[2]
                    if '/' in address:
                        address, _ = address.split('/', 1)
                    ips.append(netaddr.IPAddress(address))
            except IndexError:
                pass
    return ips


def execute_ifconfig(*ifconfig_args, **execute_params):
    command = ('/sbin/ifconfig',) + ifconfig_args
    result = sh.execute(command, stdin=False, stdout=True, stderr=True,
                        expect_exit_status=None, **execute_params)
    if result.exit_status or not result.stdout:
        raise IfconfigError(error=result.stderr)
    return result.stdout



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import sys

import pbr.version


version_info = pbr.version.VersionInfo('tobiko')

release = version_info.release_string()
version = version_info.version_string()


if __name__ == '__main__':
    message = ("release: {release}\n"
               "version: {version}\n").format(
                   release=release, version=version)
    sys.stdout.write(message)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

from heatclient.common import template_utils
import yaml

import tobiko
from tobiko.openstack import heat
from tobiko.tests.unit import openstack


class HeatTemplateFileTest(openstack.OpenstackTest):

    template_dirs = [os.path.dirname(__file__)]
    template_file = 'my-stack.yaml'

    def test_heat_template_file(self, template_file=None, template_dirs=None):
        template_file = template_file or self.template_file
        template_dirs = template_dirs or self.template_dirs
        template = heat.heat_template_file(template_file=template_file,
                                           template_dirs=template_dirs)
        self.assertIsInstance(template, heat.HeatTemplateFileFixture)
        self.assertEqual(template_file, template.template_file)
        self.assertEqual({}, template.template)
        self.assertEqual({}, template.template_files)

        tobiko.setup_fixture(template)
        template_files, template_dict = template_utils.get_template_contents(
            template_file=os.path.join(os.path.dirname(__file__),
                                       template_file))
        self.assertEqual(template_dict, template.template)
        self.assertEqual(template_files, template.template_files)
        template_yaml = yaml.safe_dump(template_dict)
        self.assertEqual(template_yaml, template.template_yaml)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.shell import ssh
from tobiko.shell import sh

CONF = config.CONF

LOG = log.getLogger(__name__)


def undercloud_ssh_client() -> ssh.SSHClientFixture:
    host_config = undercloud_host_config()
    if not host_config.hostname:
        raise NoSuchUndercloudHostname('No such undercloud hostname')
    return ssh.ssh_client(host=host_config.hostname, host_config=host_config)


class NoSuchUndercloudHostname(tobiko.TobikoException):
    message = "Undercloud hostname not specified"


class InvalidRCFile(tobiko.TobikoException):
    message = "Invalid RC file: {rcfile}"


def fetch_os_env(rcfile, *rcfiles) -> typing.Dict[str, str]:
    rcfiles = (rcfile,) + rcfiles
    errors = []
    for rcfile in rcfiles:
        try:
            result = sh.execute(f". {rcfile}; env | grep '^OS_'",
                                ssh_client=undercloud_ssh_client())
        except sh.ShellCommandFailed as ex:
            LOG.debug(f"Unable to get overcloud RC file '{rcfile}' content "
                      f"({ex})")
            errors.append(tobiko.exc_info)
        else:
            env = {}
            for line in result.stdout.splitlines():
                name, value = line.split('=')
                env[name] = value
            if env:
                return env
    for error in errors:
        LOG.exception(f"Unable to get overcloud RC file '{rcfile}' "
                      "content", exc_info=error)
    raise InvalidRCFile(rcfile=", ".join(rcfiles))


def load_undercloud_rcfile() -> typing.Dict[str, str]:
    return fetch_os_env(*CONF.tobiko.tripleo.undercloud_rcfile)


class UndercloudKeystoneCredentialsFixture(
        keystone.EnvironKeystoneCredentialsFixture):
    def get_environ(self) -> typing.Dict[str, str]:
        return load_undercloud_rcfile()


class HasUndercloudFixture(tobiko.SharedFixture):

    has_undercloud: typing.Optional[bool] = None

    def setup_fixture(self):
        self.has_undercloud = check_undercloud()


def check_undercloud() -> bool:
    try:
        ssh_client = undercloud_ssh_client()
    except NoSuchUndercloudHostname:
        return False
    try:
        ssh_client.connect(retry_count=1,
                           connection_attempts=1,
                           timeout=15.)
    except Exception as ex:
        LOG.debug('Unable to connect to undercloud host: %s', ex,
                  exc_info=1)
        return False
    return True


def has_undercloud() -> bool:
    return tobiko.setup_fixture(HasUndercloudFixture).has_undercloud


skip_if_missing_undercloud = tobiko.skip_unless(
    'TripleO undercloud hostname not configured', has_undercloud)


class UndecloudHostConfig(tobiko.SharedFixture):

    hostname: typing.Optional[str] = None
    port: typing.Optional[int] = None
    username: typing.Optional[str] = None
    key_filename: typing.Optional[str] = None

    def __init__(self, **kwargs):
        super(UndecloudHostConfig, self).__init__()
        self._connect_parameters = ssh.gather_ssh_connect_parameters(**kwargs)

    def setup_fixture(self):
        self.hostname = CONF.tobiko.tripleo.undercloud_ssh_hostname.strip()
        self.port = CONF.tobiko.tripleo.undercloud_ssh_port
        self.username = CONF.tobiko.tripleo.undercloud_ssh_username
        self.key_filename = CONF.tobiko.tripleo.undercloud_ssh_key_filename

    @property
    def connect_parameters(self):
        parameters = ssh.gather_ssh_connect_parameters(self)
        parameters.update(self._connect_parameters)
        return parameters


def undercloud_host_config() -> UndecloudHostConfig:
    return tobiko.setup_fixture(UndecloudHostConfig)


def undercloud_keystone_client():
    session = undercloud_keystone_session()
    return keystone.get_keystone_client(session=session)


def undercloud_keystone_session():
    return keystone.get_keystone_session(
        credentials=UndercloudKeystoneCredentialsFixture)


def undercloud_keystone_credentials():
    return tobiko.setup_fixture(
        UndercloudKeystoneCredentialsFixture).credentials



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


GROUP_NAME = 'keystone'
OPTIONS = [
    cfg.IntOpt('api_version',
               default=None,
               help="Identity API version"),
    cfg.StrOpt('auth_url',
               default=None,
               help="Identity service URL"),
    cfg.StrOpt('username',
               default=None,
               help="Username"),
    cfg.StrOpt('project_name',
               default=None,
               help="Project name"),
    cfg.StrOpt('password',
               default=None,
               help="Password"),
    cfg.StrOpt('domain_name',
               default=None,
               help="Domain name"),
    cfg.StrOpt('user_domain_name',
               default=None,
               help="User domain name"),
    cfg.StrOpt('project_domain_name',
               default=None,
               help="Project domain name"),
    cfg.StrOpt('project_domain_id',
               default=None,
               help="Project domain ID"),
    cfg.StrOpt('trust_id',
               default=None,
               help="Trust ID for trust scoping."),
    cfg.StrOpt('cloud_name',
               default=None,
               help=("Cloud name used pick authentication parameters from "
                     "clouds.*")),
    cfg.ListOpt('clouds_file_dirs',
                default=['.', '~/.config/openstack', '/etc/openstack'],
                help=("Directories where to look for clouds files")),
    cfg.ListOpt('clouds_file_names',
                default=['clouds.yaml', 'clouds.yml', 'clouds.json'],
                help="Clouds file names")]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import shutil
import typing

from oslo_log import log
import paramiko

from tobiko.shell import ssh


LOG = log.getLogger(__name__)

SSHClientType = typing.Union[None, ssh.SSHClientFixture, bool]


def sftp_client(ssh_client: SSHClientType) \
        -> typing.Optional[paramiko.SFTPClient]:
    if ssh_client is None:
        ssh_client = ssh.ssh_proxy_client()
    if isinstance(ssh_client, ssh.SSHClientFixture):
        return ssh_client.connect().open_sftp()
    assert ssh_client is None
    return None


def put_file(local_file: str,
             remote_file: str,
             ssh_client: SSHClientType = None):
    sftp = sftp_client(ssh_client)
    if sftp is None:
        LOG.debug(f"Copy local file: '{local_file}' -> '{remote_file}' ...")
        shutil.copyfile(local_file, remote_file)
    else:
        LOG.debug(f"Put remote file: '{local_file}' -> '{remote_file}' ...")
        with sftp:
            sftp.put(local_file, remote_file)


def get_file(remote_file: str,
             local_file: str,
             ssh_client: SSHClientType = None):
    sftp = sftp_client(ssh_client)
    if sftp is None:
        LOG.debug(f"Copy local file: '{remote_file}' -> '{local_file}' ...")
        shutil.copyfile(remote_file, local_file)
    else:
        LOG.debug(f"Get remote file: '{remote_file}' -> '{local_file}' ...")
        with sftp:
            sftp.get(remote_file, local_file)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from keystoneauth1 import loading
from keystoneauth1 import session as _session
from keystoneauth1 import plugin as _plugin
from oslo_log import log

import tobiko
from tobiko.openstack.keystone import _credentials
from tobiko import http


LOG = log.getLogger(__name__)


class KeystoneSessionFixture(tobiko.SharedFixture):

    session: typing.Optional[_session.Session] = None
    credentials: _credentials.KeystoneCredentialsType = None

    VALID_CREDENTIALS_TYPES = (_credentials.KeystoneCredentials,
                               _credentials.KeystoneCredentialsFixture,
                               type, str)

    def __init__(self,
                 credentials: _credentials.KeystoneCredentialsType = None,
                 session: typing.Optional[_session.Session] = None):
        super(KeystoneSessionFixture, self).__init__()
        if credentials is not None:
            tobiko.check_valid_type(credentials, *self.VALID_CREDENTIALS_TYPES)
            self.credentials = credentials
        if session is not None:
            self.session = session

    def setup_fixture(self):
        self.setup_session()

    def setup_session(self):
        session = self.session
        if session is None:
            credentials = _credentials.get_keystone_credentials(
                self.credentials)

            LOG.debug("Create Keystone session from credentials "
                      f"{credentials}")
            credentials.validate()
            loader = loading.get_plugin_loader('password')
            params = credentials.to_dict()
            # api version parameter is not accepted
            params.pop('api_version', None)
            params.pop('cacert', None)
            auth = loader.load_from_options(**params)
            self.session = session = _session.Session(auth=auth, verify=False)
            http.setup_http_session(session)
            self.credentials = credentials


KeystoneSessionType = typing.Union[None,
                                   _session.Session,
                                   typing.Type,
                                   str,
                                   KeystoneSessionFixture]


def keystone_session(obj: KeystoneSessionType) -> _session.Session:
    if obj is None:
        return default_keystone_session()
    if tobiko.is_fixture(obj):
        obj = tobiko.get_fixture(obj)
        if isinstance(obj, KeystoneSessionFixture):
            obj = tobiko.setup_fixture(obj).session
    if isinstance(obj, _session.Session):
        return obj
    raise TypeError(f"Can't get {_session.Session} object from {obj}")


InitSessionType = typing.Optional[typing.Callable]


class KeystoneSessionManager(object):

    def __init__(self):
        self.sessions: typing.Dict[typing.Any,
                                   KeystoneSessionFixture] = {}

    def get_session(self,
                    credentials: typing.Any = None,
                    init_session: InitSessionType = None,
                    shared: bool = True) \
            -> KeystoneSessionFixture:
        if shared:
            shared_key, session = self.get_shared_session(credentials)
        else:
            shared_key = session = None
        if session is None:
            return self.create_session(credentials=credentials,
                                       init_session=init_session,
                                       shared=shared,
                                       shared_key=shared_key)
        else:
            return session

    def get_shared_session(self, credentials: typing.Any) \
            -> typing.Tuple[typing.Any,
                            typing.Optional[KeystoneSessionFixture]]:
        if tobiko.is_fixture(credentials):
            key = tobiko.get_fixture_name(credentials)
        else:
            key = credentials
        return key, self.sessions.get(key)

    def create_session(self,
                       credentials: typing.Any = None,
                       init_session: InitSessionType = None,
                       shared: bool = True,
                       shared_key: typing.Any = None) \
            -> KeystoneSessionFixture:
        if init_session is None:
            init_session = KeystoneSessionFixture
        assert callable(init_session)
        LOG.debug('Initialize Keystone session: %r(credentials=%r)',
                  init_session, credentials)
        session: KeystoneSessionFixture = init_session(
            credentials=credentials)
        tobiko.check_valid_type(session, KeystoneSessionFixture)
        if shared:
            self.sessions[shared_key] = session
        return session


SESSIONS = KeystoneSessionManager()


def default_keystone_session(
        shared: bool = True,
        init_session: InitSessionType = None,
        manager: typing.Optional[KeystoneSessionManager] = None) -> \
        _session.Session:
    return get_keystone_session(shared=shared, init_session=init_session,
                                manager=manager)


def get_keystone_session(
        credentials: typing.Any = None,
        shared: bool = True,
        init_session: typing.Any = None,
        manager: typing.Optional[KeystoneSessionManager] = None) -> \
        _session.Session:
    if manager is None:
        manager = SESSIONS
    session = manager.get_session(credentials=credentials, shared=shared,
                                  init_session=init_session)
    tobiko.check_valid_type(session, KeystoneSessionFixture)
    return tobiko.setup_fixture(session).session


def get_keystone_endpoint(
        session: KeystoneSessionType = None,
        auth: typing.Optional[_plugin.BaseAuthPlugin] = None,
        **kwargs) -> \
        typing.Optional[str]:
    return keystone_session(session).get_endpoint(auth=auth, **kwargs)


def get_keystone_token(
        session: KeystoneSessionType = None,
        auth: typing.Optional[_plugin.BaseAuthPlugin] = None) -> \
        typing.Optional[str]:
    return keystone_session(session).get_token(auth=auth)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import os
import tempfile

from oslo_log import log


from tobiko.common import _exception

LOG = log.getLogger(__name__)


def makedirs(name, mode=0o777, exist_ok=True):
    """Creates directory and its parents if directory doesn't exists.

    This emulates Python3 os.makedirs behavior[1]

    [1] https://docs.python.org/3/library/os.html#os.makedirs)
    """
    try:
        os.makedirs(name, mode)
    except Exception:
        if not exist_ok or not os.path.isdir(name):
            raise


@contextlib.contextmanager
def open_output_file(filename, mode='w', temp_dir=None, text=False):
    basename = os.path.basename(filename)
    dirname = os.path.dirname(filename)
    prefix, suffix = os.path.splitext(basename)
    prefix += '-'
    temp_fd, temp_filename = tempfile.mkstemp(prefix=prefix,
                                              suffix=suffix,
                                              dir=temp_dir or dirname,
                                              text=text)
    try:
        with os.fdopen(temp_fd, mode) as temp_stream:
            yield temp_stream
        os.rename(temp_filename, filename)
    finally:
        with _exception.exc_info():
            if os.path.isfile(temp_filename):
                os.remove(temp_filename)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing  # noqa

from oslo_log import log

from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.shell import files
from tobiko.shell import sh
from tobiko.tripleo import _overcloud
from tobiko.tripleo import _undercloud


LOG = log.getLogger(__name__)


class TripleoTopology(topology.OpenStackTopology):

    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'tripleo_neutron_dhcp',
        neutron.L3_AGENT:  'tripleo_neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'tripleo_neutron_ovs_agent',
        neutron.METADATA_AGENT: 'tripleo_neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'tripleo_ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'tripleo_ovn_controller'
    }

    agent_to_container_name_mappings = {
        neutron.DHCP_AGENT: 'neutron_dhcp',
        neutron.L3_AGENT:  'neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'neutron_ovs_agent',
        neutron.METADATA_AGENT: 'neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'ovn_controller'
    }

    has_containers = True

    # TODO: add more known subgrups here
    known_subgroups: typing.List[str] = ['controller', 'compute']

    # In TripleO we need to parse log files directly
    file_digger_class = files.LogFileDigger

    # This is dict which handles mapping of the log file and systemd_unit (if
    # needed) for the OpenStack services
    # Format of this dict is like below:
    # service_name: (log_filename, systemd_unit_name)
    log_names_mappings = {
        neutron.SERVER: '/var/log/containers/neutron/server.log*',
    }

    def create_node(self, name, ssh_client, **kwargs):
        return TripleoTopologyNode(topology=self,
                                   name=name,
                                   ssh_client=ssh_client,
                                   **kwargs)

    def discover_nodes(self):
        self.discover_ssh_proxy_jump_node()
        self.discover_undercloud_nodes()
        self.discover_overcloud_nodes()

    def discover_undercloud_nodes(self):
        if _undercloud.has_undercloud():
            config = _undercloud.undercloud_host_config()
            ssh_client = _undercloud.undercloud_ssh_client()
            self.add_node(address=config.hostname,
                          group='undercloud',
                          ssh_client=ssh_client)

    def discover_overcloud_nodes(self):
        if _overcloud.has_overcloud():
            for server in _overcloud.list_overcloud_nodes():
                try:
                    _overcloud.power_on_overcloud_node(server)
                except Exception:
                    LOG.exception("Error ensuring overcloud node power "
                                  "status is on")
                host_config = _overcloud.overcloud_host_config(server=server)
                ssh_client = _overcloud.overcloud_ssh_client(
                    hostname=server.name,
                    host_config=host_config)
                node = self.add_node(address=host_config.hostname,
                                     hostname=server.name,
                                     group='overcloud',
                                     ssh_client=ssh_client)
                assert isinstance(node, TripleoTopologyNode)
                node.overcloud_server = server
                self.discover_overcloud_node_subgroups(node)

    def discover_overcloud_node_subgroups(self, node):
        # set of subgroups extracted from node name
        subgroups: typing.Set[str] = set()

        # extract subgroups names from node name
        subgroups.update(subgroup
                         for subgroup in node.name.split('-')
                         if is_valid_overcloud_group_name(group_name=subgroup,
                                                          node_name=node.name))

        # add all those known subgroups names that are contained in
        # the node name (controller, compute, ...)
        subgroups.update(subgroup
                         for subgroup in self.known_subgroups
                         if subgroup in node.name)

        # bind node to discovered subgroups
        if subgroups:
            for subgroup in sorted(subgroups):
                LOG.debug("Add node '%s' to subgroup '%s'", node.name,
                          subgroup)
                self.add_node(hostname=node.name, group=subgroup)
        else:
            LOG.warning("Unable to obtain any node subgroup from node "
                        "name: '%s'", node.name)
        return subgroups


class TripleoTopologyNode(topology.OpenStackTopologyNode):

    overcloud_server: typing.Optional[nova.NovaServer] = None

    def power_on_overcloud_node(self):
        server = self.overcloud_server
        if server is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is on...")
        _overcloud.power_on_overcloud_node(server)
        hostname = sh.get_hostname(ssh_client=self.ssh_client)
        LOG.debug(f"Overcloud node {self.name} power is on ("
                  f"hostname={hostname})")

    def power_off_overcloud_node(self):
        server = self.overcloud_server
        if server is None:
            raise TypeError(f"Node {self.name} is not and Overcloud server")
        self.ssh_client.close()
        LOG.debug(f"Ensuring overcloud node {self.name} power is off...")
        _overcloud.power_off_overcloud_node(server)
        LOG.debug(f"Overcloud server node {self.name} power is off.")


def is_valid_overcloud_group_name(group_name: str, node_name: str = None):
    if not group_name:
        return False
    if group_name in ['overcloud', node_name]:
        return False
    if is_number(group_name):
        return False
    return True


def is_number(text: str):
    try:
        float(text)
    except ValueError:
        return False
    else:
        return True


def setup_tripleo_topology():
    if _undercloud.has_undercloud() or _overcloud.has_overcloud():
        topology.set_default_openstack_topology_class(TripleoTopology)


def get_ip_to_nodes_dict(openstack_nodes=None):
    if not openstack_nodes:
        openstack_nodes = topology.list_openstack_nodes(group='overcloud')
    ip_to_nodes_dict = {str(node.public_ip): node.name for node in
                        openstack_nodes}
    return ip_to_nodes_dict


def str_is_not_ip(check_str):
    letters = re.compile('[A-Za-z]')
    return bool(letters.match(check_str))


def ip_to_hostname(oc_ip):
    return get_ip_to_nodes_dict()[oc_ip]


def actual_node_groups(groups):
    """return only existing node groups"""
    return set(groups).intersection(topology.list_openstack_node_groups())


def get_node(node_name):
    node_name = node_name.split('.')[0]
    return [node for node in topology.list_openstack_nodes() if
            node.name == node_name][0]


def is_composable_roles_env():
    composable_nodes = ['messaging', 'database', 'networker']
    for nodes in composable_nodes:
        if nodes in topology.list_openstack_node_groups():
            return True
    return False



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.heat import _client
from tobiko.openstack.heat import _template
from tobiko.openstack.heat import _stack

heat_client = _client.heat_client
default_heat_client = _client.default_heat_client
get_heat_client = _client.get_heat_client
HeatClient = _client.HeatClient
HeatClientFixture = _client.HeatClientFixture
HeatClientType = _client.HeatClientType

heat_template = _template.heat_template
heat_template_file = _template.heat_template_file
HeatTemplateFixture = _template.HeatTemplateFixture
HeatTemplateFileFixture = _template.HeatTemplateFileFixture

HeatStackFixture = _stack.HeatStackFixture
heat_stack_parameters = _stack.heat_stack_parameters
INIT_IN_PROGRESS = _stack.INIT_IN_PROGRESS
INIT_COMPLETE = _stack.INIT_COMPLETE
CREATE_IN_PROGRESS = _stack.CREATE_IN_PROGRESS
CREATE_COMPLETE = _stack.CREATE_COMPLETE
CREATE_FAILED = _stack.CREATE_FAILED
DELETE_IN_PROGRESS = _stack.DELETE_IN_PROGRESS
DELETE_COMPLETE = _stack.DELETE_COMPLETE
DELETE_FAILED = _stack.DELETE_FAILED



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools
import yaml

from tobiko import config
from tobiko import tripleo


CONF = config.CONF


@tripleo.skip_if_missing_undercloud
class InventoryFileTest(testtools.TestCase):

    @tripleo.skip_if_missing_tripleo_ansible_inventory
    def test_get_tripleo_ansible_inventory(self):
        inventory = tripleo.get_tripleo_ansible_inventory()
        self.assertIn('Undercloud', inventory)
        self.assertIn('Controller', inventory)
        self.assertIn('Compute', inventory)

    @tripleo.skip_if_missing_tripleo_ansible_inventory
    def test_get_tripleo_ansible_inventory_file(self):
        inventory_file = tripleo.get_tripleo_ansible_inventory_file()
        self.assertTrue(os.path.isfile(inventory_file))

    def test_has_tripleo_ansible_inventory(self):
        result = tripleo.has_tripleo_ansible_inventory()
        inventory_file = tripleo.get_tripleo_ansible_inventory_file()
        self.assertIs(inventory_file and os.path.isfile(inventory_file),
                      result)

    def test_read_tripleo_ansible_inventory(self):
        inventory_yaml = tripleo.read_tripleo_ansible_inventory()
        self.assertIsInstance(inventory_yaml, str)
        self.assertTrue(inventory_yaml)
        inventory = yaml.safe_load(inventory_yaml)
        self.assertIn('Undercloud', inventory)
        self.assertIn('Controller', inventory)
        self.assertIn('Compute', inventory)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import logging
import os
import shlex
import subprocess
import sys
import tempfile


LOG = logging.getLogger(__name__)
BASH_EXECUTABLE = '/bin/bash'
PYTHON_EXECUTABLE = os.environ.get('PYTHON', sys.executable)


def get_logger(name):
    module = sys.modules.get(name)
    if module:
        name = name_from_path(module.__file__)
    return logging.getLogger(name)


def setup_logging(main_script=None, level=logging.DEBUG) -> object:
    main_script = main_script or sys.modules['__main__'].__file__
    logging.basicConfig(
        level=level,
        stream=sys.stderr,
        format='%(name)-s: %(levelname)-7s %(asctime)-15s | %(message)s')
    return logging.getLogger(name=name_from_path(main_script))


def name_from_path(path):
    return os.path.splitext(os.path.basename(path))[0]


def normalize_path(path):
    return os.path.realpath(os.path.expanduser(path))


def execute(command, *args, **kwargs):
    capture_stdout = kwargs.pop('capture_stdout', True)
    universal_newlines = kwargs.pop('universal_newlines', True)
    check = kwargs.pop('check', True)
    shell = kwargs.pop('shell', '/bin/bash')
    environ = kwargs.pop('environ', os.environ)

    if args or kwargs:
        command = command.format(*args, **kwargs)
    command = command.strip()

    stdout = capture_stdout and subprocess.PIPE or None

    if shell:
        command_line = [shell, '-x', '-c', command]
    else:
        command_line = shlex.split(command)

    LOG.info(f"Execute: {command_line}")
    result = subprocess.run(command_line,
                            stdout=stdout, shell=False,
                            universal_newlines=universal_newlines,
                            env=dict(environ))
    if check:
        result.check_returncode()
    return result.stdout


def execute_python(command, *args, **kwargs):
    capture_stdout = kwargs.pop('capture_stdout', True)
    universal_newlines = kwargs.pop('universal_newlines', True)
    check = kwargs.pop('check', True)
    interpreter = kwargs.pop('interpreter', PYTHON_EXECUTABLE)

    if args or kwargs:
        command = command.format(*args, **kwargs)
    command = command.strip()
    command_line = [interpreter] + shlex.split(command)

    stdout = capture_stdout and subprocess.PIPE or None
    LOG.info(f"Execute: {command_line}")
    result = subprocess.run(command_line,
                            stdout=stdout, shell=False,
                            universal_newlines=universal_newlines)
    if check:
        result.check_returncode()
    return result.stdout


def get_posargs(args=None):
    if args is None:
        args = sys.argv[1:]
    return ' '.join(shlex.quote(s) for s in args)


def make_temp(*args, **kwargs):
    fd, filename = tempfile.mkstemp(*args, **kwargs)
    os.close(fd)
    return filename


def make_dir(dirname):
    if os.path.isdir(dirname):
        return False
    else:
        LOG.debug("Create directory: '%s'", dirname)
        os.makedirs(dirname)
        return True


def remove_file(filename):
    if os.path.isfile(filename):
        LOG.debug("Remove file: '%s'", filename)
        os.unlink(filename)
        return True
    else:
        return False


@contextlib.contextmanager
def stash_dir(*target_dirs: str):
    stashed_dirs = []
    for target_dir in target_dirs:
        if os.path.isdir(target_dir):
            stashed_dir = target_dir + '.stash'
            LOG.info(f"Renaming directory: {target_dir} -> {stashed_dir}")
            os.rename(target_dir, stashed_dir)
            stashed_dirs.append((stashed_dir, target_dir))

    yield

    for stashed_dir, target_dir in stashed_dirs:
        if os.path.isdir(stashed_dir):
            LOG.info(f"Restoring directory: {stashed_dir} -> {target_dir}")
            os.rename(stashed_dir, target_dir)
